/* $Id: fa_request.c,v 1.81 2001/09/27 17:33:54 jm Exp $
 * Foreign Agent - registration request handling
 *
 * Dynamic hierarchial IP tunnel
 * Copyright (C) 1998-2001, Dynamics group
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation. See README and COPYING for
 * more details.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <syslog.h>
#include <assert.h>
#include <string.h>
#include <errno.h>

#include "fa.h"
#include "fa_hash.h"
#include "fa_mn_addr.h"
#include "tunnel.h"
#include "debug.h"
#include "auth.h"
#include "dyn_ip.h"
#include "authorized.h"
#include "util.h"
#include "agent_utils.h"
#include "md5_mac.h"
#include "list.h"

extern struct fa_config *config;
extern struct sockaddr_in upper_fa_addr;
extern struct interface_entry *up_interface;
extern struct msg_key *fa_public_key;
extern struct hashtable *tunnels_hash;
extern struct bindingtable *bindings_hash;
extern struct binding_counters bcounters;
extern struct challenge_ext **last_challenges;

/* Check if enough memory and other resources to
   accept new registration request
   returns:
      0 = if enough resources
      1 = if not enough resources
*/
static int
check_resources(void)
{
	if (bcounters.bindingcount >= config->max_bindings) {
		LOG2(LOG_ALERT, "check_resources: too many concurrent "
		     "bindings - refusing new ones\n");
		return 1;
	}

	/* If needed, some other checks might be added here */
	return 0;
}


static int
coaddr_ok(struct in_addr coaddr)
{
	struct interface_entry *iface;
	struct node *node;

	if (coaddr.s_addr == config->highest_fa_addr.s_addr)
		return 1;

	for (node = list_get_first(&config->interfaces); node != NULL;
	     node = list_get_next(node)) {
		iface = (struct interface_entry *) node;
		if (iface->addr.s_addr == coaddr.s_addr)
			return 1;
	}

	return 0;
}


/* returns:
 *   0 = auth. ext. ok
 *   1 = ff_auth missing/invalid
 *   2 = mf_auth missing/invalid
 */
static int
check_auth_ext(struct in_addr addr, struct msg_extensions *ext)
{
	struct fa_spi_entry *fa_spi;

	fa_spi = get_fa_spi(0, addr, SPI_AGENT_FA);
	if (fa_spi != NULL) {
		if (ext->ff_auth == NULL ||
		    !auth_check_vendor(AUTH_ALG_MD5, fa_spi->shared_secret,
				       fa_spi->shared_secret_len,
				       (unsigned char *) ext->req,
				       ext->ff_auth))
			return 1;
		else
			return 0;
	}

	fa_spi = get_fa_spi(0, ext->req->home_addr, SPI_AGENT_MN);
	if (fa_spi != NULL) {
		if (ext->mf_auth == NULL ||
		    !auth_check(fa_spi->alg, fa_spi->shared_secret,
				fa_spi->shared_secret_len,
				(unsigned char *) ext->req,
				ext->mf_auth))
			return 2;
		else
			return 0;
	} else if (ext->mf_auth != NULL)
		return 2;

	return 0;
}

/* returns 1 on invalid request or 0 on valid request */
static int
validate_request(struct msg_extensions *ext, struct packet_from_info *info)
{
	int error_code = 0, i;
	char dev[IFNAMSIZ], error_buf[256];
	char *error_text = NULL;
	int discard_silently = 0;

	if (ext->mh_auth == NULL) {
		error_text = "no mh_auth";
		error_code = REGREP_BAD_REQUEST_FA;
	} else if (!auth_is_protected(ext->mn_nai, ext->mh_auth)) {
		error_text = "mh_auth does not protect mn_nai";
		error_code = REGREP_BAD_REQUEST_FA;
	} else if (ext->mf_auth != NULL &&
		   !auth_is_protected(ext->mn_nai, ext->mf_auth)) {
		error_text = "mf_auth does not protect mn_nai";
		error_code = REGREP_BAD_REQUEST_FA;
	} else if (ext->mn_fa_key_req_aaa != NULL &&
		   (ext->mn_aaa_auth == NULL ||
		    !auth_is_protected(ext->mn_fa_key_req_aaa,
				       ext->mn_aaa_auth))) {
		error_text = "mn_aaa_auth does not protect mn_fa_key_req";
		error_code = REGREP_BAD_REQUEST_FA;
	} else if (ext->mn_ha_key_req_aaa != NULL &&
		   (ext->mn_aaa_auth == NULL ||
		    !auth_is_protected(ext->mn_ha_key_req_aaa,
				       ext->mn_aaa_auth)))
	{
		error_text = "mn_aaa_auth does not protect mn_ha_key_req";
		error_code = REGREP_BAD_REQUEST_FA;
	} else if (ext->challenge != NULL &&
		   (ext->mf_auth == NULL ||
		    !auth_is_protected(ext->challenge, ext->mf_auth)) &&
		   (ext->mn_aaa_auth == NULL ||
		    !auth_is_protected(ext->challenge, ext->mn_aaa_auth))) {
		/* FIX: check only in the lowest FA (?) */
		error_text = "mf_auth/mn_aaa_auth does not protect challenge";
		error_code = REGREP_MN_FAILED_AUTH_FA;
		discard_silently = 1;
#ifndef DEBUG_SIMULATED_TEST
	} else if (dyn_ip_route_get(ext->req->ha_addr, dev, IFNAMSIZ) == 0 &&
		   strcmp(dev, "lo") == 0) {
		error_text = "HA address points to local interface - "
			"possible loop";
		error_code = REGREP_UNKNOWN_HA_HA; /* RFC 2002: 3.7.2 */
#endif
	} else if ((ext->req->opts & REGREQ_RESERVED) != 0) {
		error_text = "reserved option set";
		error_code = REGREP_BAD_REQUEST_FA;
	} else if (ext->encaps_del &&
		   (ext->req->opts & REGREQ_REVERSE_TUNNEL) == 0) {
		error_text = "encapsulating delivery, "
			"but no reverse tunneling\n";
		error_code = REGREP_BAD_REQUEST_FA;
	} else if ((ext->req->opts & REGREQ_VJ_HC) != 0) {
		error_text = "unsupported van Jacobson compression option set";
		error_code = REGREP_VJ_UNAVAIL_FA;
	} else if ((ext->req->opts & (REGREQ_GRE_ENCAPS |
				      REGREQ_MINIMAL_ENCAPS)) != 0) {
		error_text = "unavailable encapsulation";
		error_code = REGREP_ENCAP_UNAVAIL_FA;
	} else if ((ext->req->opts & REGREQ_MN_DECAPS) == 0 &&
		   !config->enable_fa_decapsulation &&
		   is_sender_mobile(ext)) {
		/* check this only if this FA is the lowest FA */
		error_text = "FA decapsulation denied";
		error_code = REGREP_ENCAP_UNAVAIL_FA;
	} else if ((ext->req->opts & REGREQ_REVERSE_TUNNEL) != 0 &&
		   config->enable_reverse_tunneling == 0) {
		error_text = "reverse tunneling not allowed";
		error_code = REGREP_REVERSE_TUNNEL_UNAVAIL_FA;
	} else if ((ext->req->opts & REGREQ_REVERSE_TUNNEL) == 0 &&
		   config->enable_reverse_tunneling != 0 &&
		   config->enable_triangle_tunneling == 0 &&
		   config->force_reverse_tunneling == 0) {
		error_text = "denied request for triangle tunnel - "
		     "reverse tunnel required";
		error_code = REGREP_REVERSE_TUNNEL_MANDATORY_FA;
	} else if ((ext->req->opts & REGREQ_REVERSE_TUNNEL) == 0 &&
		   config->enable_triangle_tunneling == 0) {
		error_text = "denied request for triangle tunnel";
		error_code = REGREP_ADMIN_PROHIBITED_FA;
	} else if (!authorized_check(config->authorized_networks,
			      info->src.sin_addr)) {
		error_text = "request from an unauthorized address";
		error_code = REGREP_ADMIN_PROHIBITED_FA;
	} else if (is_sender_mobile(ext) &&
		   config->allow_mobile_nodes == FALSE) {
		error_text = "MNs not allowed in this FA";
		error_code = REGREP_ADMIN_PROHIBITED_FA;
	} else if ((ext->req->opts & REGREQ_MN_DECAPS) == 0 &&
		   !coaddr_ok(ext->req->co_addr)) {
		error_text = "MN trying to use invalid care-of address";
		error_code = REGREP_INVALID_CAREOF_FA;
	} else if (info->ttl < 255 && is_sender_mobile(ext) &&
		   (config->reg_ttl_check == TTL_CHECK_ALL ||
		    (config->reg_ttl_check == TTL_CHECK_REVERSE &&
		     (ext->req->opts & REGREQ_REVERSE_TUNNEL) != 0))) {
		error_text = "MN too distant - TTL < 255";
		error_code = REGREP_REVERSE_TUNNEL_MN_TOO_DISTANT_FA;
	} else if ((i = check_auth_ext(info->src.sin_addr, ext)) != 0) {
		if (i == 2) {
			error_text = "invalid mf_auth";
			error_code = REGREP_MN_FAILED_AUTH_FA;
		} else {
			/* actually this is FA-FA auth. failure, but there is
			 * not currently any error code for it */
			error_text = "invalid ff_auth";
			error_code = REGREP_REASON_UNSPEC_FA;
		}
	} else if (ext->double_auth_ext > 0) {
		error_text = "duplicate authentication extension";
		if (ext->double_auth_ext & DOUBLE_MF_AUTH)
			error_code = REGREP_MN_FAILED_AUTH_FA;
		else
			error_code = REGREP_BAD_REQUEST_FA;
	} else if (ext->sk_auth && ext->fa_nai &&
		   (char*) ext->sk_auth < (char*) ext->fa_nai) {
		error_text = "SK_AUTH before FA_NAI";
		error_code = REGREP_REASON_UNSPEC_FA;
	} else if (ext->sk_auth && ext->prev_fa_nai &&
		   (char*) ext->sk_auth < (char*) ext->prev_fa_nai) {
		error_text = "SK_AUTH before PREV_FA_NAI";
		error_code = REGREP_REASON_UNSPEC_FA;
	} else if (ext->ff_auth == NULL && ext->fa_nai &&
		   (GET_NAI_LEN(ext->fa_nai) != config->fa_nai_len ||
		    memcmp(MSG_NAI_DATA(ext->fa_nai), config->fa_nai,
			   config->fa_nai_len) != 0))
	{
		/* lowest FA, but the FA NAI does not match */
		error_text = "unknown FA NAI";
		error_code = REGREP_REASON_UNSPEC_FA;
	} else if (ntohs(ext->req->lifetime) >
		   config->fa_default_tunnel_lifetime) {
		snprintf(error_buf, sizeof(error_buf),
			 "requested lifetime too long (%i > %i)",
			 ntohs(ext->req->lifetime),
			 config->fa_default_tunnel_lifetime);
		error_text = error_buf;
		error_code = REGREP_LONG_LIFETIME_FA;
	} else if (ext->req->home_addr.s_addr == 0) {
		error_text = "non-zero homeaddr required";
		error_code = REGREP_NONZERO_HOMEADDR_REQD_FA;
	} else if (config->enable_challenge_response &&
		   ext->challenge == NULL &&
		   (config->require_challenge || ext->mf_auth == NULL)) {
		/* FIX: check only in the lowest FA (?) */
		error_text = "missing challenge";
		error_code = REGREP_MISSING_CHALLENGE_FA;
	}

	if (error_code != 0) {
		char mn[30], coa[30], ha[30];
		dynamics_strlcpy(mn, inet_ntoa(ext->req->home_addr), 30);
		dynamics_strlcpy(coa, inet_ntoa(ext->req->co_addr), 30);
		dynamics_strlcpy(ha, inet_ntoa(ext->req->ha_addr), 30);
		if (error_text != NULL) {
			LOG2(LOG_WARNING, "SRC=%s, MN=%s, HA=%s, COA=%s - "
			     "request denied: %s\n",
			     inet_ntoa(info->src.sin_addr), mn,
			     ha, coa, error_text);
			report_discarded_msg(ext->start, ext->len, &info->src,
					     error_text);
		}
		if (!discard_silently)
			send_failure_reply(error_code, ext, info, NULL, 0);
		else
			DEBUG(DEBUG_FLAG, "Discarding reg.req. silently\n");

		return 1;
	}

	return 0;
}


static int
add_key_request(struct fa_spi_entry *fa_spi, struct msg_extensions *ext,
		int *len, char *msg)
{
	struct msg_key *key;

	/* request the session key
	 * use shared secret with highest mobility agent if one exists or
	 * public key cryptography if no security association is available */
	if (fa_spi != NULL) {
		if (*len + sizeof(struct msg_key) > MAXMSG)
			return 1;
		key = (struct msg_key *) (msg + *len);
		init_key_extension(key, VENDOR_EXT_DYNAMICS_FA_KEYREQ,
				   htonl(fa_spi->spi), 0);
		*len += GET_KEY_EXT_LEN(key);
		DEBUG(DEBUG_FLAG, " * adding fa_keyreq (len ==> %i)\n", *len);
	} else {
		/* add own public key after the copied parts */
		if (*len + GET_KEY_EXT_LEN(fa_public_key) > MAXMSG)
			return 1;
		memcpy(msg + *len, fa_public_key,
		       GET_KEY_EXT_LEN(fa_public_key));
		*len += GET_KEY_EXT_LEN(fa_public_key);
		DEBUG(DEBUG_FLAG, " * adding fa_public_key (len ==> %i)\n",
		      *len);
	}

	return 0;
}


int
add_fa_auth_ext(struct in_addr lower_addr, struct in_addr home_addr,
		unsigned char *data, unsigned char *pos)
{
	struct fa_spi_entry *fa_spi;

	/* Add the FA->FA or FA->MN authentication extension if a security
	 * association is configured */
	fa_spi = get_fa_spi(0, lower_addr, SPI_AGENT_FA);
	if (fa_spi != NULL) {
		return auth_add_vendor(
			AUTH_ALG_MD5, fa_spi->shared_secret,
			fa_spi->shared_secret_len, data,
			(struct vendor_msg_auth *) pos,
			VENDOR_EXT_DYNAMICS_FF_AUTH,
			htonl(fa_spi->spi));
	} else {
		fa_spi = get_fa_spi(0, home_addr, SPI_AGENT_MN);
		if (fa_spi != NULL) {
			return auth_add(
				fa_spi->alg, fa_spi->shared_secret,
				fa_spi->shared_secret_len, data,
				(struct msg_auth *) pos,
				MF_AUTH, htonl(fa_spi->spi));
		}
	}
	return 0;
}


#ifdef USE_TEARDOWN
/* Send tear down message to lower FA.
   returns:
       0 if succeeded sending
       1 if not successful
*/
static int
send_tear_down(struct bindingentry *binding, struct in_addr lower_addr,
	       unsigned short lower_port, struct interface_entry *iface)
{
	unsigned char msg[MAXMSG];
	unsigned char *msgpos;
	struct reg_rep *reply;
	int msglen, result;
	struct tunnel_data *t_data;
	struct registration_ext_dynamics *dyn_ext;
	struct fa_spi_entry *fa_spi;

	t_data = binding->data;
	msgpos = msg;
	msglen = 0;
	result = 0;

	/* create the reply extension */
	reply = (struct reg_rep *) msg;
	memset(reply, 0, sizeof(struct reg_rep));
	reply->lifetime = htons(0);
	reply->type = REG_REP;
	reply->code = REGREP_ACCEPTED;
	memcpy(&reply->home_addr, &binding->mn_addr, sizeof(struct in_addr));
	memcpy(&reply->ha_addr, &binding->ha_addr, sizeof(struct in_addr));
	reply->id[0] = htonl(binding->id[0]);
	reply->id[1] = htonl(binding->id[1]);
	msgpos += sizeof(struct reg_rep);

	/* add Dynamics options extension */
	dyn_ext = (struct registration_ext_dynamics *) msgpos;
	dyn_ext->type = VENDOR_EXT_TYPE2;
	dyn_ext->length = sizeof(struct registration_ext_dynamics) - 2;
	dyn_ext->reserved = 0;
	dyn_ext->vendor_id = htonl(VENDOR_ID_DYNAMICS);
	dyn_ext->sub_type = htons(VENDOR_EXT_DYNAMICS_OPTIONS);
	dyn_ext->version = VENDOR_EXT_VERSION;
	dyn_ext->opts = REG_EXT_OWN_TEAR_DOWN;
	dyn_ext->seq = 0;
	msgpos += sizeof(struct registration_ext_dynamics);

	/* add sk_auth for authentication between FAs */
	msgpos += auth_add_vendor(AUTH_ALG_MD5, binding->key, binding->keylen,
				  msg, (struct vendor_msg_auth *) msgpos,
				  VENDOR_EXT_DYNAMICS_SK_AUTH,
				  htonl(binding->spi));

	/* Add tha FA->FA auth. ext. if a security association is
	 * configured */
	fa_spi = get_fa_spi(0, lower_addr, SPI_AGENT_FA);
	if (fa_spi != NULL) {
		msgpos += auth_add_vendor(
			AUTH_ALG_MD5, fa_spi->shared_secret,
			fa_spi->shared_secret_len, msg,
			(struct vendor_msg_auth *) msgpos,
			VENDOR_EXT_DYNAMICS_FF_AUTH,
			htonl(fa_spi->spi));
	}

	msglen = msgpos - msg;
	assert(iface != NULL);
	result = own_sendto(iface->udp_sock, lower_addr, lower_port,
			    msg, msglen);
	dynamics_check_sendto(result, msglen, "send_tear_down - registration "
			      "reply");
	if (result != msglen)
		return 1;

	DEBUG(DEBUG_FLAG, "send_tear_down: sent %d bytes (of %d) to %s:%i, "
	      "dev[%s]\n", result, msglen, inet_ntoa(lower_addr),
	      ntohs(lower_port), iface->dev);

	return 0;
}
#endif


/* list structure for denial reply records */
struct denial_dst_data {
	struct in_addr home_addr;
	struct in_addr ha_addr;
	__u32 priv_ha;
	struct timeval last;
	struct denial_dst_data *next;
};
static struct denial_dst_data *denials = NULL;


/**
 * expire_denial_records:
 * @force: 1 = remove all entries, 0 = remove entries that are older than one
 * second
 *
 * Cleanup routine for denial reply records.
 */
void
expire_denial_records(int force)
{
	struct denial_dst_data *tmp, *tmp2, *prev = NULL;
	struct timeval tv;
	static time_t last_check_sec = 0;

	gettimeofday(&tv, NULL);
	/* do not expire entries more often than once per 10 seconds */
	if (!force && tv.tv_sec < last_check_sec + 10 &&
	    tv.tv_sec > last_check_sec)
		return;
	last_check_sec = tv.tv_sec;

	tmp = denials;
	while (tmp != NULL) {
		int diff = (tv.tv_sec - tmp->last.tv_sec) * 1000000 +
			tv.tv_usec - tmp->last.tv_usec;

		if (force || diff >= 1000000) {
			DEBUG(DEBUG_FLAG, "Removing expired denial reply "
			      "record for MN <%s,", inet_ntoa(tmp->home_addr));
			DEBUG(DEBUG_FLAG, "%s,%u>\n",
			      inet_ntoa(tmp->ha_addr), tmp->priv_ha);
			if (prev == NULL)
				denials = tmp->next;
			else
				prev->next = tmp->next;
			tmp2 = tmp;
			tmp = tmp->next;
			free(tmp2);
		} else {
			prev = tmp;
			tmp = tmp->next;
		}
	}
}


/**
 * send_failure_reply:
 * @code: the error code to be used in the denial reply
 * @ext: pointer to the parsed message that caused the denial reply
 * @info: source data of the request for which the denial reply is to be sent
 * @sk: pointer to the session key if available; NULL = no session key
 * @sk_len: length of the session key if @sk != NULL
 *
 * Send a failure reply to the mobile node (or the lower foreign agent if this
 * foreign agent is not the lowest one)
 *
 * Returns: 0 if successfully sent, 1 if message was not sent (error in
 * sending, more than one denial per second, or invalid destination address)
 */
int
send_failure_reply(int code, struct msg_extensions *ext,
		   struct packet_from_info *info, unsigned char *sk,
		   int sk_len)
{
	struct denial_dst_data *tmp;
	unsigned char msg[MAXMSG];
	struct reg_rep *reply;
	int msglen = 0;
	int result;
	__u32 priv_ha = 0;
	struct timeval tv;

	reply = (struct reg_rep *) msg;
	memset(reply, 0, sizeof(struct reg_rep));
	if (code == REGREP_LONG_LIFETIME_FA)
		reply->lifetime = htons(config->fa_default_tunnel_lifetime);
	else
		reply->lifetime = 0;
	reply->type = REG_REP;
	reply->code = code;
	if (ext != NULL && ext->req != NULL) {
		reply->home_addr.s_addr = ext->req->home_addr.s_addr;
		reply->ha_addr.s_addr = ext->req->ha_addr.s_addr;
		memcpy(reply->id, ext->req->id, sizeof(reply->id));
	} else if (ext != NULL && ext->rep != NULL) {
		reply->home_addr.s_addr = ext->rep->home_addr.s_addr;
		reply->ha_addr.s_addr = ext->rep->ha_addr.s_addr;
		memcpy(reply->id, ext->rep->id, sizeof(reply->id));
	}
	msglen += sizeof(struct reg_rep);

	if (ext != NULL && ext->priv_ha != NULL)
		priv_ha = ntohl(ext->priv_ha->priv_ha);

	/* check whether the last denial was sent less than a second ago */
	gettimeofday(&tv, NULL);
	tmp = denials;
	while (tmp != NULL) {
		int diff = (tv.tv_sec - tmp->last.tv_sec) * 1000000 +
			tv.tv_usec - tmp->last.tv_usec;

		if (tmp->home_addr.s_addr == reply->home_addr.s_addr &&
		    tmp->ha_addr.s_addr == reply->ha_addr.s_addr &&
		    tmp->priv_ha == priv_ha) {
			if (diff < 1000000 && diff >= 0) {
				DEBUG(DEBUG_FLAG, "More than one denial reply "
				      "per second for MN <%s,",
				      inet_ntoa(tmp->home_addr));
				DEBUG(DEBUG_FLAG, "%s,%u> - skipping this "
				      "denial\n", inet_ntoa(tmp->ha_addr),
				      tmp->priv_ha);
				return 1;
			}
			tmp->last.tv_sec = tv.tv_sec;
			tmp->last.tv_usec = tv.tv_usec;
			break;
		}

		tmp = tmp->next;
	}

	if (tmp == NULL) {
		DEBUG(DEBUG_FLAG, "Adding denial reply record for MN <%s,",
		      inet_ntoa(reply->home_addr));
		DEBUG(DEBUG_FLAG, "%s,%u>\n", inet_ntoa(reply->ha_addr),
		      priv_ha);
		tmp = (struct denial_dst_data *)
			malloc(sizeof(struct denial_dst_data));
		if (tmp == NULL) {
			DEBUG(DEBUG_FLAG, "\tcould not add record (out of "
			      "memory)\n");
			return 1;
		}
		tmp->home_addr.s_addr = reply->home_addr.s_addr;
		tmp->ha_addr.s_addr = reply->ha_addr.s_addr;
		tmp->priv_ha = priv_ha;
		tmp->last.tv_sec = tv.tv_sec;
		tmp->last.tv_usec = tv.tv_usec;
		tmp->next = denials;
		denials = tmp;
	}

	DEBUG(DEBUG_FLAG, "Sending failure reply code %i to %s:%i\n",
	      code, inet_ntoa(info->src.sin_addr), ntohs(info->src.sin_port));

#if 0
	/* FIX: currently challenge extension adding to denial replies is
	 * commented out do to possibly missing binding (see below) and
	 * ambiguous information in RFC 3012 */
	if (config->enable_challenge_response &&
	    config->challenge_in_reg_reply) {
		struct challenge_ext *challenge;
		/* FIX: in hierarchical FA network only the lowest FA should
		 * add the challenge extension */
		challenge = create_challenge_ext(config, MN_FA_CHALLENGE_EXT);
		if (challenge != NULL) {
			int n = GET_CHALLENGE_EXT_LEN(challenge);
			if (msglen + n > MAXMSG) {
				free(challenge);
				return 1;
			}
			memcpy(msg + msglen, challenge, n);
			msglen += n;
			/* FIX: this challenge must be recorded for the MN,
			 * but there may not be a binding for it yet.. */
		}
	}
#endif

	if (sk != NULL &&
	    msglen + MAX_SK_LEN + sizeof(struct vendor_msg_auth) <= MAXMSG) {
		msglen += auth_add_vendor(
			AUTH_ALG_MD5, sk, sk_len, msg,
			(struct vendor_msg_auth *) (msg + msglen),
			VENDOR_EXT_DYNAMICS_SK_AUTH,
			ext != NULL && ext->mh_auth != NULL ?
			ext->mh_auth->spi : 0);
	}

	if (msglen + MAX_SK_LEN + sizeof(struct msg_auth) <= MAXMSG)
		msglen += add_fa_auth_ext(info->src.sin_addr, reply->home_addr,
					  msg, msg + msglen);

	if (config->packet_socket_mode == PACKET_SOCKET_MODE_ONLY_RECEIVE ||
	    (ext != NULL && !is_sender_mobile(ext))) {
		if (!send_address_ok(info->src.sin_addr))
			return 1;

		result = sendto(info->iface->udp_sock, msg, msglen, 0,
				(struct sockaddr *) &info->src,
				sizeof(info->src));
	} else
		result = packet_socket_send(info, msg, msglen);

	dynamics_check_sendto(result, msglen, "send_failure_reply");

	return 0;
}


/* forward unconfirmed request or deregistration to upper FA or HA
   returns:
      0 if successful
      1 on error
*/
static int
forward_request(struct bindingentry *binding, struct msg_extensions *ext)
{
	char msg[MAXMSG];
	int len, len2, sent;
	struct fa_spi_entry *fa_spi;
	struct tunnel_data *t_data;
	char *auth_ext;
	char *copy_end; /* pointer to the ext following the copied area */
	int auth_len;
	const char *auth_txt;
	t_data = binding->data;

	DEBUG(DEBUG_FLAG, "Forwarding request upwards\n");

	if (ext->req == NULL || ext->mh_auth == NULL) {
		LOG2(LOG_WARNING, "forward_request: not req or mh_auth\n");
		return 1;
	}

	/* copy the MAC protected area directly from the received message */
	if (ext->mn_aaa_auth != NULL) {
		auth_ext = (char *) ext->mn_aaa_auth;
		auth_len = GET_GEN_AUTH_EXT_LEN(ext->mn_aaa_auth);
		auth_txt = "mn_aaa_auth";
	} else {
		auth_ext = (char *) ext->mh_auth;
		auth_len = GET_AUTH_EXT_LEN(ext->mh_auth);
		auth_txt = "mh_auth";
	}
	assert(auth_ext > (char *) ext->req);
	len = (__u8 *) auth_ext - (__u8 *) ext->req + auth_len;
	if (len < sizeof(struct reg_req) + auth_len || len > MAXMSG) {
		LOG2(LOG_WARNING, "forward_request: invalid length %i\n",
		     len);
		return 1;
	}
	memcpy(msg, ext->req, len);
	copy_end = (char *) ext->req + len;
	DEBUG(DEBUG_FLAG, " * copying up to and including %s (len ==> %i)\n",
	      auth_txt, len);

	if (ext->sk_auth && !config->highest_FA) {
		/* if we came here, this FA could not check the session key
		 * based authentication - so, just copy the data to the upper
		 * mobility agent */
		if ((void *) ext->sk_auth <= (void *) ext->mh_auth) {
			LOG2(LOG_WARNING, "forward_request: sk_auth before "
			     "mh_auth - dropping message\n");
			return 1;
		}

		len2 = (char *) ext->sk_auth - (char *) ext->mh_auth -
			GET_AUTH_EXT_LEN(ext->mh_auth) +
			GET_VENDOR_AUTH_EXT_LEN(ext->sk_auth);
		if (len + len2 > MAXMSG || len2 < 0) {
			LOG2(LOG_WARNING, "forward_request: message overflow "
			     "(len=%i, len2=%i)\n", len, len2);
			return 1;
		}
		memcpy(msg + len, ((char *) ext->mh_auth) +
		       GET_AUTH_EXT_LEN(ext->mh_auth), len2);
		len += len2;
		copy_end += len2;
		DEBUG(DEBUG_FLAG, " * copying up to and including sk_auth "
		      "(len ==> %i)\n", len);
	}

	if (config->highest_FA)
		fa_spi = get_fa_spi(0, ext->req->ha_addr, SPI_AGENT_HA);
	else
		fa_spi = get_fa_spi(0, upper_fa_addr.sin_addr, SPI_AGENT_FA);
	/* FIX: add key request even to deregistration messages if this FA is
	 * not the highest FA
	 * The keyreq is needed to know whether the lower end is MN or FA */
	if (ntohs(ext->req->lifetime) != 0 || !config->highest_FA) {
		if (add_key_request(fa_spi, ext, &len, msg) != 0)
			return 1;
	}

	/* add GRE key ext. if it is needed (duplicate private address) */
	if (t_data->up_type == TUNNEL_GRE) {
		struct gre_key_ext *gre;
		if (len + sizeof(struct gre_key_ext) > MAXMSG)
			return 1;
		gre = (struct gre_key_ext *) (msg + len);
		gre->type = VENDOR_EXT_TYPE2;
		gre->length = sizeof(struct gre_key_ext) - 2;
		gre->reserved = 0;
		gre->vendor_id = htonl(VENDOR_ID_DYNAMICS);
		gre->sub_type = htons(VENDOR_EXT_DYNAMICS_GRE_KEY);
		gre->key = htonl(t_data->up_key);
		len += sizeof(struct gre_key_ext);
		DEBUG(DEBUG_FLAG, " * adding GRE ext\n");
	}

	if (ext->challenge != NULL && (char *) ext->challenge >= copy_end) {
		int n = GET_CHALLENGE_EXT_LEN(ext->challenge);
		if (len + n > MAXMSG)
			return 1;
		memcpy(msg + len, ext->challenge, n);
		len += n;
		DEBUG(DEBUG_FLAG, " * adding Challenge ext (len=%i)\n", n);
	}

	/* Add the FA->FA or FA->HA authentication extension if a security
	 * association is configured and send the message */
	if (config->highest_FA) {
		if (fa_spi != NULL) {
			len += auth_add(
				fa_spi->alg, fa_spi->shared_secret,
				fa_spi->shared_secret_len,
				(unsigned char *) msg,
				(struct msg_auth *) (msg + len),
				FH_AUTH, htonl(fa_spi->spi));
		}
		DEBUG(DEBUG_FLAG, "forward_request ==> HA %s:%i\n",
		      inet_ntoa(ext->req->ha_addr), config->ha_udp_port);
		sent = own_sendto(up_interface->udp_sock, ext->req->ha_addr,
				  htons(config->ha_udp_port), msg, len);
	} else {
		if (fa_spi != NULL) {
			len += auth_add_vendor(
				AUTH_ALG_MD5, fa_spi->shared_secret,
				fa_spi->shared_secret_len,
				(unsigned char *) msg,
				(struct vendor_msg_auth *) (msg + len),
				VENDOR_EXT_DYNAMICS_FF_AUTH,
				htonl(fa_spi->spi));
		}
		DEBUG(DEBUG_FLAG,
		      "forward_request ==> upper FA %s:%i\n",
		      inet_ntoa(upper_fa_addr.sin_addr),
		      ntohs(upper_fa_addr.sin_port));
		if (!send_address_ok(upper_fa_addr.sin_addr))
			return 1;
		sent = sendto(up_interface->udp_sock, msg, len, 0,
			      (struct sockaddr *) &upper_fa_addr,
			      sizeof(upper_fa_addr));
	}

	assert(len < MAXMSG);

	dynamics_check_sendto(sent, len, "forward_request");
	if (sent != len)
		return 1;

	return 0;
}


/* reply to authenticated request
   returns:
      0 if successful
      1 on error
*/
static int
reply_message(struct bindingentry *binding, struct msg_extensions *ext)
{
	struct tunnel_data *t_data;
	unsigned char msg[MAXMSG];
	int len, ret, timeout;
	struct reg_rep *reply;
	struct msg_key *key;
	struct fa_spi_entry *fa_spi;
	time_t now;

	t_data = binding->data;

	DEBUG(DEBUG_FLAG, "Reply to authenticated request\n");

	memset(msg, 0, MAXMSG);
	len = 0;

	reply = (struct reg_rep *) msg;
	reply->type = REG_REP;
	reply->code = REGREP_ACCEPTED;
	time(&now);
	timeout = binding->exp_time - now;
	if (timeout < 0)
		timeout = 0;
	if (timeout > ntohs(ext->req->lifetime)) {
		/* FIX: should this kind of request be forward upwards to
		 * lower the binding in all the agents? */
		DEBUG(DEBUG_FLAG, "Request lifetime=%i < timeout=%i\n",
		      ntohs(ext->req->lifetime), timeout);
		timeout = ntohs(ext->req->lifetime);
	}
	DEBUG(DEBUG_FLAG, "\tbinding->timeout=%i, calculated timeout=%i\n",
	      binding->timeout, timeout);
	reply->lifetime = htons(timeout);
	reply->home_addr.s_addr = ext->req->home_addr.s_addr;
	reply->ha_addr.s_addr = ext->req->ha_addr.s_addr;
	reply->id[0] = ext->req->id[0];
	reply->id[1] = ext->req->id[1];
	len += sizeof(struct reg_rep);

	if (config->send_sfa_debug_ext) {
		struct sfa_debug_ext *sfa_ext = (struct sfa_debug_ext *)
			(msg + len);
		sfa_ext->type = VENDOR_EXT_TYPE2;
		sfa_ext->reserved = 0;
		sfa_ext->vendor_id = htonl(VENDOR_ID_DYNAMICS);
		sfa_ext->sub_type = htons(VENDOR_EXT_DYNAMICS_SFA_DEBUG);
		sfa_ext->length = sizeof(struct sfa_debug_ext) - 2;
		sfa_ext->sfa_addr = up_interface->addr;
		len += GET_SFA_DEBUG_EXT_LEN(sfa_ext);
		DEBUG(DEBUG_FLAG, " * added SFA debug ext\n");
	}

	if (t_data->is_lowest_fa && config->enable_challenge_response &&
	    config->challenge_in_reg_reply) {
		struct challenge_ext *challenge;
		challenge = create_challenge_ext(config, MN_FA_CHALLENGE_EXT);
		if (challenge != NULL) {
			int n = GET_CHALLENGE_EXT_LEN(challenge);
			if (len + n > MAXMSG) {
				free(challenge);
				return 1;
			}
			memcpy(msg + len, challenge, n);
			len += n;
			/* save the last challenge so that MN's registration
			 * requests can be checked to contain a valid challenge
			 */
			if (t_data->last_challenge != NULL)
				free(t_data->last_challenge);
			t_data->last_challenge = challenge;
			DEBUG(DEBUG_FLAG, " * added challenge ext\n");
		}
	}

	if (len + sizeof(struct msg_auth) + MD5_MAC_LEN > MAXMSG)
		return 1;
	len += auth_add_vendor(
		AUTH_ALG_MD5, binding->key, binding->keylen, msg,
		(struct vendor_msg_auth *) (msg + len),
		VENDOR_EXT_DYNAMICS_SK_AUTH,
		ext->sk_auth != NULL ? ext->sk_auth->spi : 0);

	/* Foreign Agent key reply - add only if not lowest FA
	 * Use shared secret and MD5 if fa_keyreq was used or RSA if
	 * fa_pubkey was used */
	if (binding->fa_spi != 0) {
		fa_spi = get_fa_spi(binding->fa_spi, binding->lower_addr,
				    SPI_AGENT_FA);
		if (fa_spi == NULL) {
			LOG2(LOG_WARNING, "unknown FA SPI %i, IP=%s\n",
			     binding->fa_spi,
			     inet_ntoa(binding->lower_addr));
			return -1;
		}
		key = (struct msg_key *) (msg + len);
		if (MSG_KEY_DATA(key) + MAX_SK_LEN > msg + MAXMSG) {
			LOG2(LOG_WARNING, "not enough space for SK to FA\n");
			return -1;
		}
		len += auth_encrypt(fa_spi->alg, fa_spi->shared_secret,
				    fa_spi->shared_secret_len, binding->key,
				    key, reply, VENDOR_EXT_DYNAMICS_FA_KEYREP,
				    htonl(fa_spi->spi));
	} else if (t_data->fa_pubkey != NULL) {
		if (t_data->last_sent_fa_pubkeyrep == NULL) {
			t_data->last_sent_fa_pubkeyrep =
			dynamics_do_rsa_encrypt(binding->key, binding->keylen,
						t_data->fa_pubkey);
			if (t_data->last_sent_fa_pubkeyrep == NULL)
				return 1;
		}

		if (len + GET_KEY_EXT_LEN(t_data->last_sent_fa_pubkeyrep) >
		    MAXMSG) {
			LOG2(LOG_WARNING, "Message buffer overflow\n");
			return 1;
		}
		memcpy(msg + len, t_data->last_sent_fa_pubkeyrep,
		       GET_KEY_EXT_LEN(t_data->last_sent_fa_pubkeyrep));
		len += GET_KEY_EXT_LEN(t_data->last_sent_fa_pubkeyrep);
	}

	len += add_fa_auth_ext(binding->lower_addr, ext->req->home_addr, msg,
			       msg + len);

	DEBUG(DEBUG_FLAG, " * sending %i bytes to %s:%i\n",
	      len, inet_ntoa(binding->lower_addr),
	      ntohs(binding->lower_port));
	assert(t_data->info.iface != NULL);
	if (config->packet_socket_mode != PACKET_SOCKET_MODE_ONLY_RECEIVE &&
	    t_data->is_lowest_fa)
		ret = packet_socket_send(&t_data->info, msg, len);
	else
		ret = own_sendto(t_data->info.iface->udp_sock,
				 binding->lower_addr,
				 binding->lower_port, msg, len);
	dynamics_check_sendto(ret, len, "reply_message");

	return 0;
}

/**
 * update_request_binding_info:
 * @binding:
 * @ext:
 * @iface: NULL if reply
 * @info: Source of the request message
 * @dst_addr: Destination address of the request message
 *
 * 
 */
static void
update_request_binding_info(struct bindingentry *binding, 
			    struct msg_extensions *ext,
			    struct packet_from_info *info)
{
	struct tunnel_data *t_data = (struct tunnel_data *) binding->data;

	/* update down interface pointer */
	memcpy(&t_data->info, info, sizeof(*info));
	t_data->old_if_index = info->iface->if_index;
	/* update lower end type (MN/FA) */
	t_data->is_lowest_fa = is_sender_mobile(ext);
	/* update MN/FA decapsulation mode */
	t_data->fa_decapsulation = t_data->is_lowest_fa &&
		(ext->req->opts & REGREQ_MN_DECAPS) == 0;
	/* update encapsulating delivery mode */
	t_data->encaps_delivery = !t_data->is_lowest_fa ||
		ext->encaps_del != NULL;


	/* update down_type and down_key */
	if (ext->gre_key != NULL) {
		t_data->down_type = TUNNEL_GRE;
		t_data->down_key = ntohl(ext->gre_key->key);
	} else if (!t_data->fa_decapsulation || t_data->encaps_delivery) {
		t_data->down_type = TUNNEL_IPIP;
		t_data->down_key = 0;
	} else {
		t_data->down_type = TUNNEL_NO_ENCAPS;
		t_data->down_key = info->iface->if_index;
	}

	/* update FA public key */
	if (!equal_pubkey(t_data->fa_pubkey, ext->fa_pubkey)) {
		if (t_data->last_sent_fa_pubkeyrep != NULL) {
			free(t_data->last_sent_fa_pubkeyrep);
			t_data->last_sent_fa_pubkeyrep = NULL;
		}
		if (t_data->fa_pubkey != NULL) {
			DEBUG(DEBUG_FLAG,
			      "Removing the previous fa_pubkey\n");
			free(t_data->fa_pubkey);
			t_data->fa_pubkey = NULL;
		}
		if (ext->fa_pubkey) {
			DEBUG(DEBUG_FLAG, "Adding new fa_pubkey\n");
			t_data->fa_pubkey = (struct msg_key *)
				malloc(GET_KEY_EXT_LEN(ext->fa_pubkey));
			if (t_data->fa_pubkey != NULL) {
				memcpy(t_data->fa_pubkey, ext->fa_pubkey,
				       GET_KEY_EXT_LEN(ext->fa_pubkey));
			} else {
				LOG2(LOG_WARNING, "Not enough memory "
				     "for fa_pubkey\n");
			}
		}
	}

	/* update binding lower address and port */
	binding->lower_addr = info->src.sin_addr;
	binding->lower_port = info->src.sin_port;

	/* update FA keyreq SPI */
	if (ext->fa_keyreq)
		binding->fa_spi = ntohl(ext->fa_keyreq->spi);
	else
		binding->fa_spi = 0;

	setup_device_forcing(t_data);

	/* update tunnel mode */
	if ((ext->req->opts & REGREQ_REVERSE_TUNNEL) != 0 || 
	    config->force_reverse_tunneling)
		t_data->tunnel_mode = TUNNEL_MODE_REVERSE;
	else
		t_data->tunnel_mode = TUNNEL_MODE_TRIANGLE;
	
#ifdef USE_TEARDOWN
	/* FIX-JM: update t_data->upper_id ?? */
#endif
}


/**
 * save_binding_information:
 * @save: Where to copy binding info
 * @binding: A binding containing the info to copy
 *
 * Copies information from @binding to @save.
 */
void
save_binding_info(struct saved_binding_info *save,
		  const struct bindingentry *binding)
{
	struct tunnel_data *t_data;

	t_data = binding->data;

	save->lower_addr = binding->lower_addr;
	save->lower_port = binding->lower_port;
	save->is_lowest_fa = t_data->is_lowest_fa;
	save->down_type = t_data->down_type;
	save->down_key = t_data->down_key;
	save->fa_decapsulation = t_data->fa_decapsulation;
	save->encaps_delivery = t_data->encaps_delivery;
	save->down_tunl = t_data->down_tunl;
	save->down_iface = t_data->info.iface;
	memcpy(save->force_route_dev, t_data->force_route_dev, IFNAMSIZ);
	memcpy(save->force_reverse_dev, t_data->force_reverse_dev, IFNAMSIZ);
	save->tunnel_mode = t_data->tunnel_mode;

	return;
}

/**
 * debug_compare_binding_info:
 * @old: 
 * @binding:
 *
 * Compare old and changed binding information and print
 * debug information about it
 */
static void
debug_compare_binding_info(const struct saved_binding_info *old,
			   const struct bindingentry *binding)
{
	struct tunnel_data *t_data;

	t_data = binding->data;

	if (t_data->info.iface != old->down_iface) {
		DEBUG(DEBUG_FLAG, "\tdown interface changed %s => %s\n",
		      old->down_iface->dev, t_data->info.iface->dev);
	}

	if (t_data->is_lowest_fa && !old->is_lowest_fa)
		DEBUG(DEBUG_FLAG, "\tlower end changed FA -> MN\n");
	else if (!t_data->is_lowest_fa && old->is_lowest_fa)
		DEBUG(DEBUG_FLAG, "\tlower end changed MN -> FA\n");

	if (old->fa_decapsulation != t_data->fa_decapsulation) {
		DEBUG(DEBUG_FLAG, "\tFA decapsulation changed %i => %i\n",
		      old->fa_decapsulation, t_data->fa_decapsulation);
	}

	if (old->encaps_delivery != t_data->encaps_delivery) {
		DEBUG(DEBUG_FLAG, "\tencaps. delivery changed: %i => %i\n",
		      old->encaps_delivery, t_data->encaps_delivery);
	}

	if (t_data->down_type != old->down_type ||
	    t_data->down_key != old->down_key) {
		DEBUG(DEBUG_FLAG, "\ttunnel type changed: %i,%i => %i,%i\n",
		      old->down_type, old->down_key, t_data->down_type,
		      t_data->down_key);
	}

	if (binding->lower_addr.s_addr != old->lower_addr.s_addr) {
		DEBUG(DEBUG_FLAG, "\tlower addr changed %s => ",
		      inet_ntoa(old->lower_addr));
		DEBUG(DEBUG_FLAG, "%s\n", inet_ntoa(binding->lower_addr));
	}

	if (binding->lower_port != old->lower_port) {
		DEBUG(DEBUG_FLAG, "\tlower port changed %i => %i\n",
		      ntohs(old->lower_port), ntohs(binding->lower_port));
	}

	if (strcmp(t_data->force_route_dev, old->force_route_dev) != 0) {
		DEBUG(DEBUG_FLAG, "\tforce route dev changed %s => %s\n",
		      old->force_route_dev, t_data->force_route_dev);
	}

	if (strcmp(t_data->force_reverse_dev, old->force_reverse_dev) != 0) {
		DEBUG(DEBUG_FLAG, "\tforce reverse dev changed %s => %s\n",
		      old->force_reverse_dev, t_data->force_reverse_dev);
	}
}
	
/**
 * handle_lower_switch:
 * @binding: MN's binding entry
 * @old: saved binding data
 * @ext: request/reply message data
 * @info: source of the request message
 *
 * Handle registration request as switched FA.
 *
 * 1. Send reply
 * 2. Make tunnel to new lower FA
 * 3. Connect the new tunnel to upper agent and change route from HA
 *    to new lower FA/MN
 * 4. Send Tear Down message
 * 5. Unconnect old tunnel
 * 6. Delete old tunnel (delayed)
 *
 * Returns: 0 if successful,
 *          1 on error
 */
int
handle_lower_switch(struct bindingentry *binding, 
		    struct saved_binding_info *old,
		    struct msg_extensions *ext,
		    struct in_addr *addr, int request_auth_ok)
{
	struct tunnel_data *t_data;
	int reverse;

	assert(binding != NULL && binding->data != NULL);

	DEBUG(DEBUG_FLAG, "Lower FA or MN address changed (locupd)\n");
	t_data = binding->data;
	assert(t_data->info.iface != NULL);

	/* tell about the changes made */
	debug_compare_binding_info(old, binding);

#ifdef DYNAMICS_SET_MAIN_TABLE_MN_ROUTE
	if (old->fa_decapsulation && !t_data->fa_decapsulation &&
	    dyn_ip_route_del(binding->mn_addr, old->down_iface->dev) != 0) {
		LOG2(LOG_WARNING, "dyn_ip_route_del(%s,%s) failed\n",
		     inet_ntoa(binding->mn_addr), old->down_iface->dev);
	}

	/* add/replace host route to MN if in FA decaps mode */
	if (t_data->is_lowest_fa && !old->is_lowest_fa &&
	    t_data->fa_decapsulation &&
	    dyn_ip_route_replace(binding->mn_addr, t_data->info.iface->dev) ==
	    -1) {
		LOG2(LOG_WARNING, "handle_lower_switch: "
		     "dyn_ip_route_replace(%s,%s) failed.\n",
		     inet_ntoa(binding->mn_addr), t_data->info.iface->dev);
	}
#endif /* DYNAMICS_SET_MAIN_TABLE_MN_ROUTE */

	/* 1. Send reply */
	/* Send reply if the request is valid
  	 * (if not valid, the reply is forwarded elsewhere) */
	if (request_auth_ok) {
		reply_message(binding, ext);
	}

	/* 2. Make tunnel to new lower FA */
	t_data->down_tunl = tunnel_add(
		tunnels_hash, *addr, binding->tun_dev,
		t_data->info.iface->addr, 1, t_data->down_type,
		t_data->down_key);
	if (t_data->down_tunl == NULL) {
		LOG2(LOG_WARNING, "handle_lower_switch: tunnel_add failed!\n");
		return 1;
	}

	/* 3. Connect the new tunnel to upper agent */
	/* Note: "To MN" routing table does not change, so no need to update
	 * routing rule for locally generated packets. */
	/* Use delayed tunnel unconnect and make a new connect */
	reverse = old->tunnel_mode == TUNNEL_MODE_REVERSE ? 1 : 0;
	if (tunnel_unconnect(t_data->up_tunl, old->down_tunl, binding->mn_addr,
			     reverse, 0, old->force_route_dev,
			     old->force_reverse_dev) < 0) {
		LOG2(LOG_WARNING,
		     "handle_lower_switch: tunnel_unconnect failed!\n");
	}

	reverse = t_data->tunnel_mode == TUNNEL_MODE_REVERSE ? 1 : 0;
	if (tunnel_connect(t_data->up_tunl, t_data->down_tunl,
			   binding->mn_addr, reverse, t_data->force_route_dev,
			   t_data->force_reverse_dev)
	    < 0) {
		LOG2(LOG_WARNING,
		     "handle_lower_switch: tunnel_connect failed!\n");
		return 1;
	}

#ifdef USE_TEARDOWN
	/* 4. Send Tear Down message */
	/* Send a tear down message only if this is not the lowest FA */
	if (old->lowest_fa == FALSE) {
		if (send_tear_down(binding, old->lower_addr, old->lower_port,
				   old->down_iface))
			DEBUG(DEBUG_FLAG, "send_tear_down failed\n");
	} else
		DEBUG(DEBUG_FLAG, "Lowest FA - no tear down reply sent\n");
#endif

	/* 5. Delayed tunnel deletion (6. delete tunnel) */
	if (tunnel_delete_ptr(old->down_tunl, 0) < 0) {
		LOG2(LOG_WARNING, "handle_lower_switch: tunnel delete "
		     "failed\n");
		return 1;
	}

	return 0;
}


/* Create tunnel and table for the tunnel. The table will get route for MN.
 * This enables data coming from upper mobility agent to be routed to MN.
 * --> Lesser packet loss when SFA or HA changes the route (data path is
 * already made available for downstream packets). 
 * 
 * For security reasons, do nothing if this is the HFA.
 * 
 * Returns 0 if successful, else 1 */
static int
make_preliminary_data_path(struct bindingentry *binding,
			   struct unconfirmed_request *unc,
			   struct interface_entry *iface)
{
	struct tunnel_data *t_data;
	struct tunnel *down, *up;

	/* Highest FA MUST NOT do this optimization or there may be some
	 * security problems (data transfered before getting a reply from the
	 * HA) */
	if (config->highest_FA)
		return 0;

	t_data = binding->data;

	/* tunnel downwards (if FA decaps is used, this will be just a dummy
	 * entry for the non-capsulated connection) */
	down = tunnel_add(tunnels_hash, unc->info.src.sin_addr,
			  binding->tun_dev, iface->addr, 1, unc->tunnel_type,
			  unc->tunnel_key);
	if (down == NULL) {
		LOG2(LOG_WARNING, "make_preliminary_data_path: "
		     "tunnel_add failed (down)\n");
		return 1;
	}

	up = tunnel_add(tunnels_hash, upper_fa_addr.sin_addr,
			t_data->ha_tun_dev, up_interface->addr, 1,
			t_data->up_type, t_data->up_key);
	if (up == NULL) {
		LOG2(LOG_WARNING, "make_preliminary_data_path: "
		     "tunnel_add failed (up)\n");
		tunnel_delete_ptr(down, 0);
		return 1;
	}

	/* FIX-JM: what about force_route_dev? should it be != NULL when
	 * FA decaps is used with encaps. del.? */
	/* FIX-JM: what about force_reverse_dev? should it be != NULL when
	 * MN decaps is used without encaps. del.? */
	/* Route for HA->MN direction - this enables routing of downstream
	 * packets to the MN. Note that this does not make the reverse
	 * connection. */
	if (tunnel_connect(up, down, binding->mn_addr, 0, NULL, NULL) < 0) {
		LOG2(LOG_WARNING, "make_preliminary_data_path: "
		     "tunnel_connect failed\n");
	}

	/* With lazy tunnel deletion the tunnel will be up long enough.
	 * If no reply is received, the tunnel is removed automatically.
	 * Additionally the reference count will be ok. When reply is
	 * received the tunnels are remade which means that the reference
	 * counts are increased. */
	if (tunnel_unconnect(up, down, binding->mn_addr, 0, 0, NULL, NULL) < 0)
	{
		LOG2(LOG_WARNING, "make_preliminary_data_path: "
		     "tunnel_unconnect failed\n");
	}

	if (tunnel_delete_ptr(down, 0) == -1) {
		LOG2(LOG_WARNING, "make_preliminary_data_path: "
		     "lazy tunnel delete failed (down).\n");
	}
	if (tunnel_delete_ptr(up, 0) == -1) {
		LOG2(LOG_WARNING, "make_preliminary_data_path: "
		     "lazy tunnel delete failed (up).\n");
	}

	return 0;
}


static int check_mn_challenge(struct tunnel_data *t_data,
			      struct msg_extensions *ext)
{
	int i, found = 0;

	/* check whether the MN has already used this challenge; if yes and
	 * this is not a replay of request with same id, send denial reg. reply
	 * with code REGREP_STALE_CHALLENGE_FA */
	for (i = 0; i < 2 * config->challenge_window; i++) {
		if (equal_challenge(t_data->used_challenges[i],
				    ext->challenge)) {
			found = 1;
			break;
		}
	}
	if (found && (i != t_data->used_challenges_pos ||
		      memcmp(ext->req->id, t_data->used_challenges_id,
			     sizeof(ext->req->id)) != 0))
		return REGREP_STALE_CHALLENGE_FA;

	/* check whether the used challenge is known */
	found = 0;
	for (i = 0; i < config->challenge_window; i++) {
		if (equal_challenge(last_challenges[i], ext->challenge)) {
			found = 1;
			break;
		}
	}
	if (!found && equal_challenge(t_data->last_challenge,
				      ext->challenge)) {
		/* let the MN use the same challenge only once by freeing the
		 * stored challenge */
		free(t_data->last_challenge);
		t_data->last_challenge = NULL;
		found = 1;
	}
	if (!found)
		return REGREP_UNKNOWN_CHALLENGE_FA;

	return 0;
}


/**
 * handle_request:
 * @ext: the request with extensions
 * @info: source of the request
 * @dst_addr: destination address of the request
 * @iface: interface that the request arrived on
 * @config: configuration file parameters
 * @ttl: TTL of the request
 *
 * Handle registration request.
 *
 * Returns:
 *     0 if successful,
 *     1 on error
 */
int
handle_request(struct msg_extensions *ext, struct packet_from_info *info,
	       struct fa_config *config)
{
	struct bindingentry *binding;
	int auth_ok, ret = 0;
	int lower_mode_changed = 0;
	struct tunnel_data *t_data;
	struct unconfirmed_request *unc = NULL;
	struct bindingkey bkey;
	struct interface_entry *prev_iface;
	int iface_ok;

	DEBUG(DEBUG_FLAG, "Handling request from MN %s",
	      inet_ntoa(ext->req->home_addr));
	DEBUG(DEBUG_FLAG, " (%s:%d)\n", inet_ntoa(info->src.sin_addr),
	      ntohs(info->src.sin_port));

	if (validate_request(ext, info))
		return 1;

	memcpy(&bkey.mn_addr, &(ext->req->home_addr), sizeof(bkey.mn_addr));
	memcpy(&bkey.ha_addr, &(ext->req->ha_addr), sizeof(bkey.ha_addr));
	if (ext->priv_ha != NULL)
		bkey.priv_ha = ntohl(ext->priv_ha->priv_ha);
	else
		bkey.priv_ha = 0;
	binding = binding_fetch(bindings_hash, &bkey);
	if (binding == NULL) {
		int type, added_addr = 0;
		DEBUG(DEBUG_FLAG, "No binding for MN => making new binding\n");

		if (config->max_pending > 0 &&
		    bcounters.pendingcount >= config->max_pending) {
			LOG2(LOG_ALERT, "Too many pending registration "
			     "requests - refusing new ones\n");
			send_failure_reply(REGREP_NO_RESOURCES_FA, ext,
					   info, NULL, 0);
			return 1;
		}

		if (config->enable_challenge_response && ext->challenge) {
			int i, found = 0;
			for (i = 0; i < config->challenge_window; i++) {
				if (equal_challenge(last_challenges[i],
						    ext->challenge)) {
					found = 1;
					break;
				}
			}
			if (!found) {
				send_failure_reply(REGREP_UNKNOWN_CHALLENGE_FA,
						   ext, info, NULL, 0);
				return 1;
			}
		}

		if (config->highest_FA && ext->priv_ha == NULL)
			type = MN_ADDR_TUNNEL_IPIP;
		else {
			type = mn_addr_add(&ext->req->home_addr);
			DEBUG(DEBUG_FLAG, "mn_addr_add(%s) => %i\n",
			      inet_ntoa(ext->req->home_addr), type);
			added_addr = 1;
		}
		if (type != MN_ADDR_TUNNEL_ERROR) {
			int max_time = config->fa_default_tunnel_lifetime;
			if (config->delete_pending_after > 0 &&
			    config->delete_pending_after < max_time)
				max_time = config->delete_pending_after;
			binding = make_new_binding(
				bindings_hash, ext, max_time, &bcounters);
		}
		if (binding == NULL) {
			LOG2(LOG_WARNING, type > -2 ?
			     "Could not make new binding\n" :
			     "Out of resources while checking MN address");
			send_failure_reply(REGREP_NO_RESOURCES_FA, ext,
					   info, NULL, 0);
			if (added_addr)
				mn_addr_remove(&ext->req->home_addr, type);

			return 1;
		}
		t_data = (struct tunnel_data *) binding->data;
		if (type == MN_ADDR_TUNNEL_IPIP) {
			t_data->up_type = TUNNEL_IPIP;
			t_data->up_key = 0;
		} else {
			t_data->up_type = TUNNEL_GRE;
			t_data->up_key = type;
		}
		if (added_addr)
			t_data->mn_addr_added = 1;
	} else {
		check_unconfirmed_timeout(binding);
		t_data = (struct tunnel_data *) binding->data;

		if (config->enable_challenge_response && ext->challenge) {
			int ret = check_mn_challenge(t_data, ext);
			if (ret != 0) {
				send_failure_reply(ret, ext, info, NULL, 0);
				return 1;
			}
		}
	}

	if (ext->challenge) {
		int n = t_data->used_challenges_pos + 1;
		if (n >= 2 * config->challenge_window)
			n = 0;
		t_data->used_challenges_pos = n;
		if (t_data->used_challenges[n] != NULL)
			free(t_data->used_challenges[n]);
		t_data->used_challenges[n] = (struct challenge_ext *)
			malloc(GET_CHALLENGE_EXT_LEN(ext->challenge));
		if (t_data->used_challenges[n] == NULL) {
			DEBUG(DEBUG_FLAG, "malloc failed for "
			      "used_challenges\n");
		} else {
			memcpy(t_data->used_challenges[n], ext->challenge,
			       GET_CHALLENGE_EXT_LEN(ext->challenge));
		}
	}

	auth_ok = FALSE;

	if (!t_data->confirmed) {
		DEBUG(DEBUG_FLAG, "\tunconfirmed binding\n");
		if (check_resources()) {
			LOG2(LOG_ALERT,
			     "FA: Not enough resources to accept "
			     "registration\n");
			send_failure_reply(REGREP_NO_RESOURCES_FA, ext,
					   info, NULL, 0);
			return 1;
		}
		unc = add_unconfirmed_data(bindings_hash, &bcounters,
					   binding, ext, info, config);
	} else {
		DEBUG(DEBUG_FLAG, "\tconfirmed binding\n");
		if (ext->sk_auth != NULL) {
#ifdef DEBUG_SIMULATED_TEST
			/* for debuging only - accept any sk_auth extension */
			if (0) {
			}
#else
			if (!auth_check_vendor(AUTH_ALG_MD5, binding->key,
					       binding->keylen,
					       (unsigned char *) ext->req,
					       ext->sk_auth)) {
				DEBUG(DEBUG_FLAG,
				      "handle_request: sk_auth not valid\n");
			}
#endif
			else if (ext->ext_dyn == NULL) {
				DEBUG(DEBUG_FLAG,
				      "handle_request: missing ext_dyn\n");
			} else if ((void *) ext->ext_dyn >
				   (void *) ext->sk_auth) {
				DEBUG(DEBUG_FLAG, "handle_request: ext_dyn "
				      "after sk_auth\n");
			} else if (ntohl(ext->ext_dyn->seq) <=
				   t_data->last_used_seq_num) {
				LOG2(LOG_WARNING, "handle_request: seq# did "
				     "not increase\n");
			} else {
				auth_ok = TRUE;
				t_data->last_used_seq_num =
					ntohl(ext->ext_dyn->seq);
			}
		}
	}

	/* check, whether this FA is the switching FA */
	iface_ok = 0;
	if (ext->prev_fa_nai) {
		prev_iface = fa_hash_check(ext->prev_fa_nai);
		if (GET_NAI_LEN(ext->prev_fa_nai) == config->fa_nai_len &&
		    memcmp(MSG_NAI_DATA(ext->prev_fa_nai), config->fa_nai,
			   config->fa_nai_len) == 0) {
			iface_ok = 1;
			DEBUG(DEBUG_FLAG,
			      "This FA was the previous FA - okay to reply\n");
		} else if (prev_iface == NULL) {
			DEBUG(DEBUG_FLAG, "Unknown previous FA NAI - "
			      "forwarding request upwards\n");
		} else {
			DEBUG(DEBUG_FLAG, "Request from a known FA below us - "
			      "okay to reply\n");
			iface_ok = 1;
		}
	} else {
		DEBUG(DEBUG_FLAG, "No FA NAI ext - using only session key "
		      "based SFA detection\n");
		iface_ok = 1;
	}

	if (ext->fa_nai) {
		prev_iface = fa_hash_check(ext->fa_nai);
		if (prev_iface != NULL && prev_iface != info->iface) {
			LOG2(LOG_ALERT, "Request coming from different "
			     "interface - forcing forwarding of the "
			     "request\n");
			LOG2(LOG_ALERT, "old iface '%s', now iface '%s'\n",
			     prev_iface->dev, info->iface->dev);
		}
	}

	/* handle authorized messages self and forward unauthorized messages
	 * and disconnect requests to upper mobility agent */
	if (auth_ok && ntohs(ext->req->lifetime) > 0 && iface_ok) {
		/* check whether FA decapsulation to MN has changed */
		if (t_data->fa_decapsulation !=
		    (is_sender_mobile(ext) &&
		     (ext->req->opts & REGREQ_MN_DECAPS) == 0)) {
			DEBUG(DEBUG_FLAG, "FA decaps changed (old=%i) => "
			      "forcing lower switch\n",
			      t_data->fa_decapsulation);
			lower_mode_changed = 1;
		}
		/* check whether MN encapsulation to FA has changed */ 
		if (t_data->encaps_delivery !=
		    (!is_sender_mobile(ext) || ext->encaps_del != NULL)) {
			DEBUG(DEBUG_FLAG, "Encaps. delivery changed (old=%i) "
			      "=> forcing lower switch\n",
			      t_data->encaps_delivery);
			lower_mode_changed = 1;
		}
		if (unc != NULL && (t_data->down_type != unc->tunnel_type ||
				    t_data->down_key != unc->tunnel_key)) {
			DEBUG(DEBUG_FLAG, "Down tunnel mode changed "
			      "(%i,%i => %i,%i) => forcing lower switch\n",
			      t_data->down_type, t_data->down_key,
			      unc->tunnel_type, unc->tunnel_key);
			lower_mode_changed = 1;
		}

		if (unc != NULL && t_data->info.iface != unc->info.iface) {
			DEBUG(DEBUG_FLAG, "Down interface changed (%s => %s) "
			      "=> forcing lower switch\n",
			      t_data->info.iface->dev, unc->info.iface->dev);
			lower_mode_changed = 1;
		}

		if (binding->lower_addr.s_addr != info->src.sin_addr.s_addr ||
		    lower_mode_changed) {
			struct saved_binding_info old;
			save_binding_info(&old, binding);
			update_request_binding_info(binding, ext, info);
			handle_lower_switch(binding, &old, ext,
					    &info->src.sin_addr, TRUE);
		}
		else
			reply_message(binding, ext);

		time(&binding->mod_time);
		binding->id[0] = ntohl(ext->req->id[0]);
		binding->id[1] = ntohl(ext->req->id[1]);
#ifdef USE_TEARDOWN
		/* Note: t_data->upper_id is left unchanged because upper
		 * FAs/HA do not get this id */
#endif
	} else {
		/* unauthorized message or disconnect request */
		int use_prelim = 0;
		if (t_data->confirmed)
			unc = add_unconfirmed_data(bindings_hash, 
						   &bcounters,
						   binding, ext, info, config);

		/* use preliminary data path optimization when there is not
		 * yet a tunnel downwards or the lower end has changed */
		if (!t_data->confirmed)
			use_prelim = 1;
		else if (auth_ok && binding->lower_addr.s_addr !=
			 info->src.sin_addr.s_addr)
			use_prelim = 1;

		if (ntohs(ext->req->lifetime) > 0 && use_prelim && unc != NULL
		    && make_preliminary_data_path(binding, unc,
						  info->iface) != 0) {
			DEBUG(DEBUG_FLAG, "preliminary data path optimization "
			      "failed\n");
			ret = 1;
		}

		t_data->pending_request = TRUE;
		/* forward request first */
		forward_request(binding, ext);
	} 
	
	return ret;
}
