/* $Id: mn_util.c,v 1.89 2001/10/20 14:20:38 jm Exp $
 * Mobile Node - help functions
 *
 * Dynamic hierarchial IP tunnel
 * Copyright (C) 1998-2001, Dynamics group
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation. See README and COPYING for
 * more details.
 * 
 * Additional modifications Copyright 2010 Juho Paaso, Aalto 
 * University / Department of Communication and Networking"
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <getopt.h>
#include <signal.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <assert.h>
#ifdef DYN_TARGET_WINDOWS
#include <fcntl.h>
#endif

#include "debug.h"
#include "dyn_ip.h"
#include "proxyarp.h"
#include "agentapi.h"
#include "util.h"
#ifdef WITH_WIRELESS
#include "monitor.h"
#endif
#include "mn_handler.h"
#include "mn.h"
#include "mn_tunnel.h"
#ifdef INCLUDE_IPAY
#include "mn_ipay.h"
#endif
#include "device_info.h"
#include <sys/ioctl.h>


extern struct mn_data mn;
extern struct mn_config config;
extern struct timeval timers[TIMER_COUNT];
extern char* program_name;
extern char *opt_config;
extern int opt_foreground;
extern int real_tunnel_up; 	/* from mn_tunnel */

extern int mn_traces;								/* for trace prints */
extern struct ro_entry ro_cache[RO_CACHE_MAX_SIZE]; /* for HAaRO */
extern int ro_entry_count;

struct nw_iface usable_nw_ifaces[NW_IFACE_ARRAY_LEN];
int usable_nw_iface_count = 0;

static int wireless_extensions;

/* get_next_usable_iface: Finds the next usable interface from current_iface
 * and places it into the next_iface. If current_iface is not known, the first
 * iface in the usable_nw_ifaces list will be placed into the next_iface. The 
 * first iface in the list has a route with the lowest metric.
 */
int get_next_usable_iface(char *current_iface, char *next_iface) 
{
	int i;
	
	DEBUG(DEBUG_INFO, "get_next_usable_iface: \n");
	//update_usable_nw_ifaces();
	
	/* Browse all nw device usable in HAaRO */
	for (i = 0; i < usable_nw_iface_count; i++) 
	{
		/* Find out which device in the array is currently in use and register through the next one */
		if (strncmp(usable_nw_ifaces[i].dev_name, current_iface, IFNAMSIZ) == 0) 
		{
			DEBUG(DEBUG_INFO, "Current device: %d. %s\n", i, current_iface);
			
			/* Increase i, unless if it's the last index */
			if (i != (usable_nw_iface_count - 1)) i++;
			else i = 0;
			
			/* Place the next iface in the next_iface */
			memcpy(next_iface, usable_nw_ifaces[i].dev_name, IFNAMSIZ);
			DEBUG(DEBUG_INFO, "Next iface: %d. %s\n", i, next_iface);	

			return 0;
		}
	}
	/* current_iface not found from the list -> place the first iface in the list */
	memcpy(next_iface, usable_nw_ifaces[0].dev_name, IFNAMSIZ);
	DEBUG(DEBUG_INFO, "current_iface not found from the list -> use the first iface: 0. %s\n", next_iface);

	return 0;
	
}
struct nw_iface *get_nw_iface_with_name(char *iface_name)
{
	int i;
	for (i = 0; i < usable_nw_iface_count; i++)
	{
		if (strncmp(usable_nw_ifaces[i].dev_name, iface_name, IFNAMSIZ) == 0)
		{
			return &usable_nw_ifaces[i];
		}
	}
	return NULL;
}

int get_dev_name_by_ip(struct in_addr addr, char *device) {
	int i;
	//update_usable_nw_ifaces();
	DEBUG(DEBUG_INFO, "get_dev_name_by_ip: Find device corresponding to address %s: ",
		inet_ntoa(addr));
	for (i = 0; i < usable_nw_iface_count; i++)
	{
		if (usable_nw_ifaces[i].ip_addr.s_addr == addr.s_addr)
		{
			memcpy(device, usable_nw_ifaces[i].dev_name, IFNAMSIZ);
			DEBUG(DEBUG_INFO, "%s\n", device);
			return 1;
		}
	}
	DEBUG(DEBUG_INFO, "Not found!\n", device);
	return 0;
}

int check_if_iface_usable(char *iface_name) {
	int i;
	//update_usable_nw_ifaces();
	for (i = 0; i < usable_nw_iface_count; i++)
	{
		if (strncmp(usable_nw_ifaces[i].dev_name, iface_name, IFNAMSIZ) == 0)
		{
			DEBUG(DEBUG_INFO, "check_if_iface_usable: Iface %s is usable\n",
				iface_name);
			return 1;
		}
	}
	DEBUG(DEBUG_INFO, "check_if_iface_usable: Iface %s is not usable!\n",
		iface_name);
	return 0;
}

/* send_udp_to: Send len bytes from buf to dst_ip:dst_port.
 * 
 * @dst_port: in nw byte order
 * 
 */ 
int send_udp_to(char *buf, int len, struct in_addr dst_ip, int dst_port, int socket) {
	
	struct sockaddr_in dst_address;
	int bytes_sent;
 	dst_address.sin_family		= AF_INET;
  	dst_address.sin_port		= dst_port;
  	dst_address.sin_addr.s_addr	= dst_ip.s_addr;
  	bzero(&(dst_address.sin_zero), 8); 

	bytes_sent = sendto(socket, buf, len, 0, (struct sockaddr *) &dst_address, 
		sizeof(struct sockaddr));
	
	if (bytes_sent < 0) {
 		LOG2(LOG_ERR, "send_udp_to - sendto: %s\n", strerror(errno));
 	}
	return 0;
}

/* Get the port number socket is binded to in nw byte order */
short get_socket_local_port(int sock) {
	struct sockaddr_in sin;
	socklen_t len = sizeof(sin);
	
	if (getsockname(sock, (struct sockaddr *)&sin, &len) == -1)
	{
	    perror("getsockname - ");
	    return -1;
	} else {
	    DEBUG(DEBUG_INFO, "get_socket_local_port: Socket %d port number %d\n", 
	    	sock, ntohs(sin.sin_port));
	    return sin.sin_port;
	}
}

/* update_usable_nw_ifaces: Fill array usable_nw_ifaces with system's networking 
 * interfaces (name and IP address) that:
 *  - are up
 *  - are running 
 *  - are not in link local network
 *  - have a default gateway
 * 
 * The array is organised so that the interface with a route with the lowest 
 * metric is in the first index etc.
 */
int update_usable_nw_ifaces() {
	int old_if_index, k, original_k, start_i, smallest_i, current_i; 
	int temp_route_metric;
	struct in_addr dev_addr, link_local_addr;
	struct nw_iface temp_nw_iface;
	/* Link local nw address 169.254.0.0/16 */
	inet_aton("169.254.0.0", &link_local_addr);
	
	DEBUG(DEBUG_INFO, "System's networking devices and addresses usable in HAaRO:\n");

	/*------------------------------------------------------------------*
	 | Find the next device in the mn.iface array.						|
	 *------------------------------------------------------------------*/
	old_if_index 		= dyn_ip_get_ifindex(mn.device_in_use);
	
	usable_nw_iface_count = 0;
	k = mn.dev_index_in_iface_array;
	original_k 	= k;
	while (1) 
	{			
		/* IP address of the device into dev_addr */
		dyn_ip_get_ifaddr(mn.iface[k].device, &dev_addr);
		temp_route_metric = get_metric_if_default_gw(mn.iface[k].device);
		
		/*------------------------------------------------------------------*
		 | Check that the next device is up etc.							|
		 *------------------------------------------------------------------*/
		if (!is_in_link_local_nw(&dev_addr) &&
			(dyn_ip_get_ifflags(mn.iface[k].device) & IFF_UP) &&
			(dyn_ip_get_ifflags(mn.iface[k].device) & IFF_RUNNING) &&
	 		temp_route_metric != -1) 
		{
			/* Copy the device name */
			memcpy(usable_nw_ifaces[usable_nw_iface_count].dev_name, 
				mn.iface[k].device, IFNAMSIZ); 
			
			/* Copy the device address */
			usable_nw_ifaces[usable_nw_iface_count].ip_addr.s_addr = dev_addr.s_addr;
			
			/* Save the metric value of the route through the device */		
			usable_nw_ifaces[usable_nw_iface_count].route_metric = temp_route_metric;
			
			/* Create a socket if it doesn't exist */
			if (usable_nw_ifaces[usable_nw_iface_count].listener_socket < 0)
			{
				usable_nw_ifaces[usable_nw_iface_count].listener_socket = 
					create_ro_socket(dev_addr);
				bind_socket_to_device(usable_nw_ifaces[usable_nw_iface_count].listener_socket, 
					usable_nw_ifaces[usable_nw_iface_count].dev_name);
			}
			
			DEBUG(DEBUG_INFO, " - %s: %s, listener socket %d\n", 
				usable_nw_ifaces[usable_nw_iface_count].dev_name,
				inet_ntoa(usable_nw_ifaces[usable_nw_iface_count].ip_addr),
				usable_nw_ifaces[usable_nw_iface_count].listener_socket);
			
			usable_nw_iface_count++;
		}
		/* Increase index k */
		if (k == mn.device_count - 1) k = 0;
		else k++;
		
		/* Check if k has gone through all the interfaces */		
		if (k == original_k) {
			break;
		}
	}	
	/*------------------------------------------------------------------*
	 | Reorganize the array by the metric value.						|
	 *------------------------------------------------------------------*/
	for (start_i = 0; start_i < usable_nw_iface_count; start_i++)
	{
		smallest_i = start_i;
		for (current_i = start_i + 1; current_i < usable_nw_iface_count; current_i++)
		{
			if (usable_nw_ifaces[current_i].route_metric < usable_nw_ifaces[smallest_i].route_metric)
				smallest_i = current_i;
		}
		/* swap if needed */
		if (smallest_i != start_i)
		{
			/* temp <- start */
			memcpy(&temp_nw_iface, &usable_nw_ifaces[start_i], sizeof(struct nw_iface));
			/* start <- smallest*/
			memcpy(&usable_nw_ifaces[start_i], &usable_nw_ifaces[smallest_i], sizeof(struct nw_iface));
			/* smallest <- temp */
			memcpy(&usable_nw_ifaces[smallest_i], &temp_nw_iface, sizeof(struct nw_iface));
		}
	}
	return 0;	
}
/* get_keygen_token: Generate a Home-Test or a Care-of-Test keygen token depending on the
 * value of the parameter is_care_of:
 *  - is_care_of == 0 
 *     --> generate Home-Test keygen token: HMAC_MD5(node_key, (HoA | nonce | 0)) 
 *  - is_care_of != 0
 *     --> generate Care-of-Test keygen token: HMAC_MD5(node_key, (CoA | nonce | 1))
 * 
 * Place the first 64 bits of the code generated by the HMAC_MD5 function in keygen_token.
 */
int get_keygen_token(__u64 *node_key, struct in_addr *sender_hoa_or_coa, 
                     __u64 *nonce, __u64 *keygen_token, int is_care_of) 
{
	unsigned int inbuf_len = 14;
	unsigned char mac_input_buf[inbuf_len];		/* 112 bit buffer */
	unsigned char full_mac_output_buf[16];		/* 128 bit buffer */
	unsigned char *key;
	unsigned int key_len;
	unsigned char *data;
	unsigned int data_len;
	unsigned char *mac;
	int offset = 0;
	unsigned char zero = 0;
	unsigned char one = 1;
	
	key			= (unsigned char *) node_key;
	key_len		= 8;	

	/* bytes 0-3: HoA or CoA */
	memcpy(mac_input_buf, &sender_hoa_or_coa->s_addr, 4);
	offset += 4;
	
	/* bytes 4-12: Nonce */
	memcpy(mac_input_buf + offset, nonce, 8);
	offset += 8;
	
	/* byte 13: 1 or 0 */
	if (is_care_of) memcpy(mac_input_buf + offset, &one, 1);
	else            memcpy(mac_input_buf + offset, &zero, 1);
  
	hmac_md5(key, key_len, mac_input_buf, inbuf_len, full_mac_output_buf);

	/* Copy the first 8 bytes from the result code buffer into the return buffer */
	memcpy(keygen_token, full_mac_output_buf, 8);
	
	return 0;
}
int get_home_keygen_token(__u64 *node_key, struct in_addr *sender_hoa, 
                          __u64 *nonce, __u64 *keygen_token) 
{
	return get_keygen_token(node_key, sender_hoa, nonce, keygen_token, 0);
}
int get_co_keygen_token(__u64 *node_key, struct in_addr *sender_coa, 
                        __u64 *nonce, __u64 *keygen_token) 
{
	return get_keygen_token(node_key, sender_coa, nonce, keygen_token, 1);
}

/* get_nonce_with_index: Get the nonce with index nonce_index.
 * 
 */
__u64 get_nonce_with_index(__u16 nonce_index) 
{
	__u16 nonce_index_difference;
	int old_nonce_array_len = NONCE_ARRAY_LEN;
	
	DEBUG(DEBUG_INFO, "get_nonce_with_index %d - ", nonce_index);
	
	if (nonce_index == mn.current_nonce_index) {
		/* requested nonce is the current nonce */
		DEBUG(DEBUG_INFO, "requested for the current nonce: %llu\n", mn.nonces[0]);
		return mn.nonces[0];
	} else if (nonce_index > mn.current_nonce_index) {
		DEBUG(DEBUG_INFO, "Error: bigger than the current nonce index %d!\n", mn.current_nonce_index);
		return 0;
	} else {
		/* requested nonce is an old nonce */
		DEBUG(DEBUG_INFO, "requested for an old nonce (current nonce index %d)\n", mn.current_nonce_index);
		nonce_index_difference = mn.current_nonce_index - nonce_index;
		if (nonce_index_difference > old_nonce_array_len) {
			DEBUG(DEBUG_INFO, "Error: Too old nonce requested! Only %d old nonces contained\n", 
				old_nonce_array_len);
		} else {
			DEBUG(DEBUG_INFO, "Returning nonces[%d]: %llu\n", nonce_index_difference, mn.nonces[nonce_index_difference]);
			return mn.nonces[nonce_index_difference];
		}
	}
}

int add_new_nonce_into_nonces_array(__u64 *nonce_to_add)
{
	int i;
	int array_len = NONCE_ARRAY_LEN;
	for(i = array_len; i >= 1; i--)
	{
		/* 5 <- 4, 4 <- 3, ... 1 <- 0 */
		mn.nonces[i] = mn.nonces[i - 1];
		 
	}
	mn.nonces[0] = *nonce_to_add;
}
int create_new_nonce()
{
	__u64 new_nonce = (__u64) rand() << 32 + (__u64) rand();
	
	/* Move old nonces one step back and add the new one */
	add_new_nonce_into_nonces_array(&new_nonce);
	mn.current_nonce_index++;
	gettimeofday(&mn.current_nonce_creation_time, NULL);
	DEBUG(DEBUG_INFO, "Created a new nonce with index %d: %llu\n", 
		mn.current_nonce_index, mn.nonces[0]);
	
	return 0;
}

/* calculate_krm: Calculate Registration management key (KRm) from the Home and 
 * Care-of token with formula: MD5(Home Keygen Token | Care-of Keygen Token).
 * 
 * @home_keygen_token: includes the 64 bit home keygen token 
 * @co_keygen_token: includes the 64 bit care-of keygen token
 * @krm: pointer to a 128 bit buffer the krm can be placed to
 * 
 */
int calculate_krm(__u64 *home_keygen_token, __u64 *co_keygen_token, __u64 *krm) {
	unsigned int inbuf_len = 16;
	unsigned char mac_input_buf[inbuf_len];

	int offset = 0;
		
	/* bytes 0-7: Home Keygen Token */
	memcpy(mac_input_buf, home_keygen_token, 8);
	offset += 8;
	
	/* bytes 8-15: Care-of Keygen Token */
	memcpy(mac_input_buf + offset, co_keygen_token, 8);
	
	/* Get the MAC code into the cr->krm */
	md5(mac_input_buf, inbuf_len, krm);
	
	return 0;
}

/* calculate_krm_for_ro_entry: Calculates the KRm (Registration Management Key) from 
 * the Home Keygen Token and the Care-of Keygen Token in this ro_entry.
 * The formula is: MD5(Home Keygen Token | Care-of Keygen Token). The resulting
 * 128 bit code is placed in the krm field in this struct ro_entry and in the mr_coas
 * array.
 */
int calculate_krm_for_ro_entry(struct ro_entry *cr) 
{
	calculate_krm(&cr->hot_keygen_token, &cr->cot_keygen_token, cr->krm);
	
	DEBUG(DEBUG_MESSAGES, "Registration management key (KRm) for %s: %llu %llu\n", 
		inet_ntoa(cr->cr_hoa), cr->krm[0], cr->krm[1]);
	
	add_krm_info_into_mr_coas_array(cr);
	
	return 0;
}

/* get_authenticator: Calculate the authenticator used in MR-CR Authentication extension.
 * 
 * Authenticator: HMAC_MD5(KRm, (The whole Reg packet until the auth field))
 * 
 * @krm: 128 bit registration management key
 * @reg_msg_start: pointer to the beginning of the Registration Request message
 * @msg_len_until_auth_field
 * @authenticator
 * 
 */
int get_authenticator(__u64 *krm, char *reg_msg_start, int msg_len_until_auth_field, __u64 *authenticator) {
	hmac_md5(krm, 16, 								/* Key */
		reg_msg_start, msg_len_until_auth_field, 	/* Input data */
		(unsigned char *) authenticator);			/* Output code */
	return 0;
}
/* check_mr_cr_auth: Verify the MR-CR Authentication extension. Check if the 
 * authenticator field corresponds to the auth_value calculated with the formula:
 * 
 * Home token    = first_64_bits( HMAC_MD5(node_key, (MR HoA | home_nonce | 0)) )
 * Care-of token = first_64_bits( HMAC_MD5(node_key, (MR CoA | co_nonce   | 1)) )
 * 
 * auth_value    = MD5(Home token | Care-of token)
 * 
 * Returns: 0 if authentication is valid, -1 if it's not.
 */
int check_mr_cr_auth(struct msg_extensions *ext, struct in_addr sender_hoa, 
			struct in_addr sender_coa, struct mr_cr_auth_ext *mr_cr_auth) 
{
	/* All in host byte order */
	__u64 authenticator_from_ext[2];		/* Authenticator from the MR-CR Auth extension */
	__u64 calculated_krm[2];				/* Registration management key calculated locally */
	__u64 calculated_authenticator[2];		/* Authenticator calculated locally */
	__u64 home_token, co_token;				/* Tokens calculated locally */
	__u64 home_nonce, co_nonce;				/* Nonces contained locally */
	__u16 home_nonce_index, co_nonce_index;	/* Nonce indexes from the extension */
	int msg_len_until_auth_ext;				/* Length from Req header start to MR-CR Auth header start */
	int msg_len_until_auth_field;
	int comparison_result;
	
	assert(ext != NULL && ext->req != NULL && ext->mr_cr_auth != NULL);
	
	/* Get the length of the message until the MR-CR Auth header's authenticator field */
	msg_len_until_auth_ext 		= (unsigned char *) ext->mr_cr_auth - (unsigned char *) ext->req;
	msg_len_until_auth_field	= msg_len_until_auth_ext + 8;
	
	DEBUG(DEBUG_MESSAGES, "check_mr_cr_auth: Lentgh from Registration header to "
		"MR-CR Auth header's auth field: %d bytes\n", msg_len_until_auth_field);
	
	/* Read the MR-CR Auth header */
	home_nonce_index 			= ntohs(mr_cr_auth->home_nonce_index);
	co_nonce_index 				= ntohs(mr_cr_auth->care_of_nonce_index);
	authenticator_from_ext[0]	= ntoh64(mr_cr_auth->authenticator[1]);		/* [43 00] <- [12 34] */
	authenticator_from_ext[1]	= ntoh64(mr_cr_auth->authenticator[0]);		/* [43 21] <- [12 34] */
	
	DEBUG(DEBUG_MESSAGES, "Read from the MR-CR Auth header: Home nonce index %u, "
		"Care-of nonce index %u, authenticator %llu %llu \n", home_nonce_index, co_nonce_index, 
		authenticator_from_ext[0], authenticator_from_ext[1]);
	
	/* Calculate tokens, KRm and authenticator from the nonces and HoA and CoA of the Reg Req sender */
	home_nonce 	= get_nonce_with_index(home_nonce_index);
	co_nonce 	= get_nonce_with_index(co_nonce_index);
	get_home_keygen_token(mn.node_key, &sender_hoa, &home_nonce, &home_token);
	get_co_keygen_token(mn.node_key, &sender_coa, &co_nonce, &co_token);
	calculate_krm(&home_token, &co_token, calculated_krm);
	get_authenticator(calculated_krm, (char *) ext->req, msg_len_until_auth_field, calculated_authenticator);

	DEBUG(DEBUG_MESSAGES, "Calculated from local values: Home token %llu, Care-of token %llu, "
		"authenticator %llu %llu\n",
		home_token, co_token, calculated_authenticator[0], calculated_authenticator[1]);
	
	comparison_result = memcmp((char *) calculated_authenticator, (char *) authenticator_from_ext, 16);
	
	if (comparison_result == 0) {
		DEBUG(DEBUG_MESSAGES, "Authenticators match\n");
		return 0;
	} else {
		DEBUG(DEBUG_MESSAGES, "Authenticators don't match!\n");
		return -1;
	}
}

/* set_timer_expiration: Set timer to expire at now + msecs_to_add
 * 
 * TODO: Check msecs_to_add max value and change the datatype
 */
int set_timer_expiration(struct timeval *timer, unsigned int msecs_to_add) 
{	
	/* Init timer to the current time */
	gettimeofday(timer, NULL);
	
	/* Add the milliseconds to the timer */
	add_usecs(timer, msecs_to_add * 1000);

	return 0;
}
/* add_seconds_to_timer: Set timer to expire at now + seconds_to_add
 */
int add_seconds_to_timer(struct timeval *timer, unsigned int seconds_to_add)
{
	/* Init timer to the current time */
	gettimeofday(timer, NULL);
	
	/* Add the seconds to the timer */
	timer->tv_sec += seconds_to_add;
	
	return 0;
}
/* handle_keepalive_reply: Handle a keepalive reply if icmp_seq found from keepalive_array. */
int handle_keepalive_reply(struct requested_ping *keepalive_array, unsigned short icmp_seq, 
			int *rtt_array, int *ping_failure_counter) 
{
	int i, rtt_in_ms;
	struct timeval now, rtt;
	
	gettimeofday(&now, NULL);
	
	/*--------------------------------------------------------------------------*
	 | Check if the reply is for one of the keepalive packets currently online 	|
	 *--------------------------------------------------------------------------*/
	for (i = 0; i < KEEPALIVE_ARRAY_LEN; i++) {
		if (keepalive_array[i].timer_on &&
			keepalive_array[i].seq_number == icmp_seq) 
		{
			/*----------------------------------------------*
			 | Set this packet's timers off				 	|
			 *----------------------------------------------*/
			keepalive_array[i].timer_on = 0;
			keepalive_array[i].failure_timer.tv_sec = 0;
			keepalive_array[i].failure_timer.tv_usec = 0;
			
			/*--------------------------------------------------------------*
			 | Calculate the RTT of this connection and add it in the array	|
			 *--------------------------------------------------------------*/
			timersub(&now, &keepalive_array[i].sending_time, &rtt);	/* rtt = now - sending_time */
			rtt_in_ms = rtt.tv_usec / 1000;
			if (rtt.tv_sec) rtt_in_ms += rtt.tv_sec * 1000;
			insert_RTT(rtt_array, rtt_in_ms);
			DEBUG(DEBUG_MESSAGES, "sent %li.%06li, ", 
				keepalive_array[i].sending_time.tv_sec, keepalive_array[i].sending_time.tv_usec);
			DEBUG(DEBUG_MESSAGES, "RTT: %ld usec\n", rtt.tv_usec);
			*ping_failure_counter = 0;
			return 0;
		}
	}
	return -1;
}
/**
 * Bind socket s to network interface iface 
 */
int bind_socket_to_device(int s, char *iface) {
	struct ifreq ifr;
	
	memset(&ifr, 0, sizeof(ifr));
	DEBUG(DEBUG_INFO, "Bind socket %d to device %s\n", s, iface);
	// Write iface into ifr.ifr_name 
	snprintf(ifr.ifr_name, sizeof(ifr.ifr_name), "%s", iface);
	
	if (setsockopt(s, SOL_SOCKET, SO_BINDTODEVICE,
		(void *)&ifr, sizeof(ifr)) < 0) 
	{
		perror("setsockopt(SO_BINDTODEVICE)");
		return -1;
	}
	return 0;
}

/**
 * Insert 4 byte UDP Tunnel Message header in *buffer
 */ 
void insert_udp_tnl_msg_header(unsigned char *buffer) {
	
	struct udp_tnl_data *tnl_data;	
	tnl_data = (struct udp_tnl_data *) buffer;
	
	tnl_data->type 			= 4;
	tnl_data->next_header 	= 4;
	tnl_data->reserved		= htons(0x00);
}

/* Returns the average RTT in milliseconds calculated from the rtt_array
 * that contains RTT_ARRAY_LEN items. However, if the average RTT is less 
 * than config.min_keepalive_interval, config.min_keepalive_interval is returned. 
 */
int get_average_RTT(int *rtt_array) {
	int i;
	int average_RTT = 0;
	int min_return = config.min_keepalive_interval;
	
	for (i = 0; i < RTT_ARRAY_LEN; i++) {
		average_RTT += rtt_array[i];
	}
	average_RTT /= 5;
	/* Won't return less than min_return */
	if (average_RTT < min_return) 
		average_RTT = min_return;
	return average_RTT;
}
/* insert_RTT: Inserts RTT into rtt_array[0] and moves the other items 
 * one step back dropping out the last element.
 */
int insert_RTT(int *rtt_array, int rtt_in_milliseconds) {
	int i;
	
	for (i = RTT_ARRAY_LEN - 1; i > 0; i--) {
		/* 4 = 3, 3 = 2, ... */
		rtt_array[i] = rtt_array[i - 1];
	}
	rtt_array[0] = rtt_in_milliseconds;
	
	return 0;
}
/* Set the RTT values to be rtt_msecs */
void init_RTTs_to_value(int *rtt_array, int array_length, int rtt_msecs) {
	int i;
	for (i = 0; i < array_length; i++) rtt_array[i] = rtt_msecs;
}

/* Set the RTT values to be the value got from the configuration file */
void init_RTTs(int *rtt_array, int array_length) {
	init_RTTs_to_value(rtt_array, array_length, config.init_rtt);
}

void init_keepalives(struct requested_ping *keepalive_array, int array_length) {
	int i;
	for (i = 0; i < array_length; i++) {
		keepalive_array[i].timer_on = 0;
	}
}

int get_iface_mip_socket(char *dev_name)
{
	int i;
	for (i = 0; i < NW_IFACE_ARRAY_LEN; i++)
	{
		if (strncmp(usable_nw_ifaces[i].dev_name, dev_name, IFNAMSIZ) == 0)
		{
			DEBUG(DEBUG_INFO, "get_iface_mip_socket: %s fd %d\n",
				dev_name, usable_nw_ifaces[i].listener_socket);
			return usable_nw_ifaces[i].listener_socket;
		}
	}
	fprintf(stderr, "Iface %s tried to use in HAaRO but it's not usable!\n", dev_name);
	return -1;
}
/* get_iface_address: Get the IP address of the parameter iface_name 
 * and place it in the parameter addr.
 */
int get_iface_address(char *iface_name, struct in_addr *addr)
{
	int fd;
	struct ifreq ifr;
	
	/* Get the IP address of the iface_name */
	fd = socket(AF_INET, SOCK_DGRAM, 0);
	ifr.ifr_addr.sa_family = AF_INET;		/* IPv4 address */
	strncpy(ifr.ifr_name, iface_name, IFNAMSIZ-1);
 	ioctl(fd, SIOCGIFADDR, &ifr);	
 	close(fd);
 
	/* Place the address in addr */
 	addr->s_addr = ((struct sockaddr_in *) &ifr.ifr_addr)->sin_addr.s_addr;

 	return 0;
}

int add_tun_fds_into_fd_set(fd_set *set) {
	int i;
	DEBUG(DEBUG_INFO, "TUN FDs into the fd_set: ");
	/* HA */
  	if (mn.tun_fd > 0) {
  		DEBUG(DEBUG_INFO, "%d (TUN FD to HA)", mn.tun_fd);
	  	FD_SET(mn.tun_fd, set);
  	}
	
	/* CRs */
	for (i = 0; i < ro_entry_count; i++) {
		if (ro_cache[i].tun_fd > 0)	{
			DEBUG(DEBUG_INFO, ", %d (%s)", 
				ro_cache[i].tun_fd, ro_cache[i].tun_dev_name);
			if (ro_cache[i].tun_fd >= 0)
				FD_SET(ro_cache[i].tun_fd, set);
		}
	}
	DEBUG(DEBUG_INFO, "\n");
	return 0;
}

int check_if_ro_fds_set(fd_set *set) {
	int i;
	
	/* HA TUNs */
  	if (mn.tun_fd > 0) {
		if (FD_ISSET(mn.tun_fd, set)) {
			DEBUG(DEBUG_INFO, "HA TUN FD is set\n");
  			handle_msg_in_ha_tun_fd();
  		}
  	}
  	
  	/* CR TUNs */
	for (i = 0; i < ro_entry_count; i++) {
		if (ro_cache[i].tun_fd > 0)	{
			if (FD_ISSET(ro_cache[i].tun_fd, set)) {
				DEBUG(DEBUG_INFO, "Msg in TUN FD %d (%s)\n", 
					ro_cache[i].tun_fd, ro_cache[i].tun_dev_name);
				handle_msg_in_cr_tun_fd(&ro_cache[i]);
			}
		}
		if (FD_ISSET(ro_cache[i].registration_socket, set)) {
			DEBUG(DEBUG_INFO, "Msg in CR (%s) Registration socket\n", inet_ntoa(ro_cache[i].cr_hoa));
			handle_cr_msg(ro_cache[i].registration_socket);
		}
		if (FD_ISSET(ro_cache[i].coti_socket, set)) {
			DEBUG(DEBUG_INFO, "Msg in CR (%s) CoTI socket\n", inet_ntoa(ro_cache[i].cr_hoa));
			handle_cr_msg(ro_cache[i].coti_socket);
		}
	}
	
	/* Listen to messages on HoA random port (HoT).	*/
	if (FD_ISSET(mn.hoa_socket, set)) {
		handle_cr_msg(mn.hoa_socket);
	}

  	/* Listen to messages on HoA or CoA port 434. */
	if (FD_ISSET(mn.hoa_listener_socket, set)) {
		handle_cr_msg(mn.hoa_listener_socket);
	}

	/* Listen to interfaces' MIP ports' sockets. */
	for (i = 0; i < NW_IFACE_ARRAY_LEN; i++) {
		if (usable_nw_ifaces[i].listener_socket > 0) {
			if (FD_ISSET(usable_nw_ifaces[i].listener_socket, set))
			{
				DEBUG(DEBUG_INFO, "Msg in iface %s MIP socket\n", usable_nw_ifaces[i].dev_name);
				handle_cr_msg(usable_nw_ifaces[i].listener_socket);
			}
		}
	}
	return 0;
}
/* create_socket: Create an UDP socket and bind it to the bind_addr:bind_port. 
 * 
 * Returns: the socket's fd.
 */
short create_socket(struct in_addr bind_addr, unsigned short bind_port) {
	int s;
	int one = 1;
	struct sockaddr_in local_addr;
	
	/* Create socket */
	s = socket(AF_INET, SOCK_DGRAM, 0);
	if (s < 0) {
		perror("create_socket: socket() -");
		return -1;
	} 
	
	DEBUG(DEBUG_INFO, "create_socket: %s:%d socket fd %d.\n", 
		inet_ntoa(bind_addr), ntohs(bind_port), s);

	/* Set IP packet info */
	if (setsockopt(s, SOL_IP, IP_PKTINFO, &one, sizeof(one)) < 0)
    	perror("setsockopt(IP_PKTINFO): ");

	/* Bind the socket to bind_addr:bind_port */
	memset((char *) &local_addr, 0, sizeof(local_addr));
	local_addr.sin_family 		= AF_INET;
	local_addr.sin_addr.s_addr 	= bind_addr.s_addr;
	local_addr.sin_port 		= bind_port;
	
	if (bind(s, (struct sockaddr *) &local_addr, sizeof(local_addr)) < 0) {
		char error[20];
		snprintf(error, 50, "create_socket: bind to %s:%d - ", 
			inet_ntoa(bind_addr), bind_port);
		perror(error);
		close(s);
		return -1;
	}
	
	return s;
}
/* create_ro_socket: Create an UDP socket and bind it to 
 * the address bind_addr (HoA or CoA) port 434.
 * 
 * Returns: the socket's fd.
 */
short create_ro_socket(struct in_addr bind_addr) {
	return create_socket(bind_addr, htons(434));
}


/**
 * check_ignore_iface:
 * @ifname: Interface name
 *
 * Check whether an interface is to be ignored
 *
 * Returns: 1 if it should be ignored, else 0
 */
static int check_ignore_iface(const char *ifname)
{
	struct ignore_iflist_entry *i;
	struct node *node;

	if (strcmp(ifname, TUNNEL_DEVICE) == 0)	
		return 1;
	if (strcmp(ifname, TUNNEL_DEVICE_B) == 0) 
		return 1;

	for (node = list_get_first(&config.ignore_iflist); node != NULL;
	     node = list_get_next(node)) {
		i = (struct ignore_iflist_entry *)node;
		if (strncmp(ifname, i->ifname, IFNAMSIZ) == 0) 
			return 1;
	}

	return 0;
}

struct fa_spi_entry* get_fa_spi(int spi, struct in_addr addr)
{
	struct fa_spi_entry *s;
	struct node *node;
	time_t now;

	time(&now);

	node = list_get_first(&config.fa_spi_list);
	while (node != NULL) {
		s = (struct fa_spi_entry *) node;
		node = list_get_next(node);

		/* Expire dynamic entries that are not infinite (lifetime=0) */
		if (s->created != 0 && s->lifetime != 0 &&
		    now > s->created + s->lifetime) {
			DEBUG(DEBUG_INFO, "get_fa_spi: dynamic "
			      "security association (SPI=%i, addr=%s) "
			      "expired\n", s->spi, inet_ntoa(s->addr));
			list_remove((struct node *) s);
			free(s);
			continue;
		}

		if ((spi != 0 && s->spi != spi) ||
		    s->addr.s_addr != addr.s_addr)
			continue;

		return s;
	}
	return NULL;
}


int add_fa_spi(struct fa_spi_entry *spi, int replace_any_spi)
{
	struct fa_spi_entry *old;

	old = get_fa_spi(replace_any_spi ? 0 : spi->spi, spi->addr);
	if (old != NULL) {
		if (old->created == 0) {
			DEBUG(DEBUG_INFO, "add_fa_spi: trying to replace "
			      "static security association - aborting "
			      "operation\n");
			return -1;
		}

		DEBUG(DEBUG_INFO, "add_fa_spi: overwriting previous dynamic "
		      "security association\n");
		remove_fa_spi(old);
		free(old);
		old = NULL;
	}

	DEBUG(DEBUG_INFO, "add_fa_spi: adding dynamic security association "
	      "(SPI=%i, addr=%s)\n", spi->spi, inet_ntoa(spi->addr));
	list_init_node(&spi->node);
	list_add_tail(&config.fa_spi_list, &spi->node);
	return 0;
}


void remove_fa_spi(struct fa_spi_entry *spi)
{
	DEBUG(DEBUG_INFO, "remove_fa_spi: removing security association "
	      "(SPI=%i, addr=%s)\n", spi->spi, inet_ntoa(spi->addr));
	list_remove(&spi->node);
}


/** 
 * send_gratuitous_arp:
 * @route_addr: 
 * @gratuituous_addr:
 *
 * A gratuituous arp message for address @gratuitous_addr will be sent 
 * on the interface where there is a route to @route_addr.
 */
void send_gratuitous_arp(struct in_addr route_addr, 
			 struct in_addr gratuitous_addr)
{
	char arp_if[IFNAMSIZ];

	if (dyn_ip_route_get(route_addr, arp_if, IFNAMSIZ) != 0) {
		DEBUG(DEBUG_INFO, "Could not get ARP interface\n");
		return;
	}
	proxyarp_gratuitous(gratuitous_addr, arp_if);
}


#ifdef MN_LOCUPD_PROFILER
void write_profiler(const char *msg)
{
	struct timeval tv;
	gettimeofday(&tv, NULL);
	fprintf(mn.profile, "%s %u.%06u (%li usec)\n", msg,
		(unsigned int) tv.tv_sec, (unsigned int) tv.tv_usec,
		(tv.tv_sec - mn.last_api.tv_sec) * 1000000 +
		tv.tv_usec - mn.last_api.tv_usec);
}
#endif


/**
 * update_fa_decsps_routes:
 * @ifname: interface name
 * @ifindex: interface index @ifname
 * @fa_addr: foreign agent addres
 * @home_net_addr: home network addres
 * @home_net_addr_plen: home network prefix length, -1 if no
 *   home network
 *
 * Set the default route to @ifname via @fa_addr and delete all
 * routes to @home_net_addr/@home_net_addr_plen.
 *
 * Returns: 0 if successful
 */
int update_fa_decaps_routes(const char *ifname, int ifindex, 
			    struct in_addr fa_addr, 
			    struct in_addr home_net_addr, 
			    int home_net_addr_plen)
{
	DEBUG(DEBUG_INFO, "FA decapsulation - setting default route to "
	      "%s via %s\n", ifname, inet_ntoa(fa_addr));

	mn.cur_route_info.known = 1;
	mn.cur_route_info.via.s_addr = fa_addr.s_addr;
	mn.cur_route_info.ifindex = ifindex;
	memcpy(mn.cur_route_info.ifname, ifname, IFNAMSIZ);
	mn.cur_route_info.ifindex_net = -1;

	if (dyn_ip_route_replace_default(ifname, &fa_addr,
					 &config.mn_home_ip_addr) != 0) {
		LOG2(LOG_ERR, "Set default route (to %s via %s) failed\n",
		     ifname, inet_ntoa(fa_addr));
		return -1;
	}
	if (home_net_addr_plen > -1) {
		int count;
		DEBUG(DEBUG_INFO, "Deleting home net route (%s/%i)\n",
		      inet_ntoa(home_net_addr), home_net_addr_plen);
		count = dyn_ip_route_delete_net_routes(&home_net_addr,
						       home_net_addr_plen);
		if (count < 0) {
			LOG2(LOG_ERR, "Home net route (%s/%i) delete failed\n",
			     inet_ntoa(home_net_addr), home_net_addr_plen);
#ifdef DYN_TARGET_WINDOWS
			/* at least Win98 does not seem to like to remove
			 * home net route.. */
			DEBUG(DEBUG_INFO, "Trying to set net route via the FA "
			      "to circumvent failed net route removal\n");

			if (dyn_ip_route_set_net(ifname, &home_net_addr,
						 home_net_addr_plen,
						 NULL, 2) < 0) {
				DEBUG(DEBUG_INFO, "Could not increase home "
				      "net route metric\n");
			}

			if (dyn_ip_route_set_net(ifname, &home_net_addr,
						 home_net_addr_plen,
						 &fa_addr, 1) < 0) {
				DEBUG(DEBUG_INFO, "Could not set net route via"
				      " FA\n");
			} else
				mn.home_net_route_set_via_fa = 1;
#endif /* DYN_TARGET_WINDOWS */
		} else {
			DEBUG(DEBUG_INFO,
			      "   %i net route%s deleted\n", count,
			      count != 1 ? "s" : "");
		}
	}

	return 0;
}

/**
 * modify_new_interface:
 * @iface: new or removed interface
 * @idx: interface map
 */
static void modify_new_interface(struct interface_data *iface, 
				 struct idxmap *idx)
{
	struct event_INTERFACE data;
	int ret;

	DEBUG(DEBUG_INFO, "modify_new_interface (start): device_count %d\n",
	      mn.device_count);
	if (config.enable_fa_decapsulation) {
		ret = dyn_ip_iface_address(idx->index, 
					   &config.mn_home_ip_addr);
		if ((ret == 2 || ret == 0) &&
		    config.use_aaa &&
		    config.mn_home_ip_addr.s_addr == 0 &&
		    config.enable_fa_decapsulation) {
			DEBUG(DEBUG_INFO, "modify_interface: interface with no"
			      " addresses - using AAA to obtain a home "
			      "address\n");
		} else {

			if (ret == 0) {
				DEBUG(DEBUG_INFO, "FA decapsulation and an "
				      "interface does not have the home "
				      "address\n");
				mn.warn_str = "FA decapsulation and an "
					"interface does not have the home "
					"address";
				return;
			}

			/* continue only if the interface has a correct
			 * address */
			if (ret != 1) {
				DEBUG(DEBUG_INFO, "modify_interface: no "
				      "addresses found (ret=%i)\n", ret);
				return;
			}
		}
	}

	if (iface->handlers_off == 1) {
		DEBUG(DEBUG_INFO, "modify_interface: interface up\n");
		memset(&data, 0, sizeof(data));
		data.iface = iface;
		data.ifname = iface->device;
		data.index = &iface->index;
		data.icmp_sock = &iface->s;
		data.icmp_sock_adv = &iface->s_adv;
		assert(idx != NULL);
		data.idx = idx;
		data.last_reg_send_time = &mn.last_reg_send_time;
		data.current_adv = &mn.current_adv;
		data.priority = &iface->priority;
		ret = handler_call_all(INTERFACE_INIT, &data);
		if (ret == -1) {
			DEBUG(DEBUG_INFO, "modify_interface: interface "
			      "initialization failed\n");
			return;
		}
		iface->handlers_off = 0;
		if (config.enable_fa_decapsulation &&
		    mn.current_adv != NULL &&
		    real_tunnel_up) {
			DEBUG(DEBUG_INFO, "modify_new_interface: fix possibly "
			      "changed routing entries\n");
			update_fa_decaps_routes(mn.current_adv->ifname,
						mn.current_adv->ifindex,
						mn.fa_addr,
						config.home_net_addr,
						config.home_net_addr_plen);
		}
	} 
	mn.device_count++;

	DEBUG(DEBUG_INFO, "modify_new_interface (end): device_count %d\n",
	      mn.device_count);

	/* go from passive find to find agent state if a new interface is up
	 * and there was previously no interfaces available OR if the current 
	 * adv. was removed */
	if ((mn.device_count == 1 && mn.state == MN_PASSIVE_FIND) /*||
	    (mn.current_adv == NULL && mn.state != MN_DISCONNECTED)*/) {
		find_agent(STATE_INIT);
	}
}

/**
 * modify_removed_interface:
 * @iface: interface
 */
static void modify_removed_interface(struct interface_data *iface)
{
	struct event_INTERFACE data;
	int ret;

	DEBUG(DEBUG_INFO, "modify_remved_interface (start): device_count %d\n",
	      mn.device_count);

	if (iface->handlers_off != 1) {
		DEBUG(DEBUG_INFO, "modify_removed_interface: interface down\n");
		memset(&data, 0, sizeof(data));
		data.ifname = iface->device;
		data.index = &iface->index;
		data.icmp_sock = &iface->s;
		data.icmp_sock_adv = &iface->s_adv;
		ret = handler_call_all(INTERFACE_DOWN, &data);
		if (ret == -1) {
			DEBUG(DEBUG_INFO, "modify_interface: interface "
			      "down failed\n");
			return;
		}
		iface->handlers_off = 1;
	} 
	mn.device_count--;
	if (mn.device_count < 0) {
		DEBUG(DEBUG_INFO, "modify_interface: device "
		      "count < 0!\n");
		mn.device_count = 0;
	}

	DEBUG(DEBUG_INFO, "modify_remived_interface (end): device_count %d\n",
	      mn.device_count);

	/* go from passive find to find agent state if the current 
	 * adv. was removed */
	if (mn.current_adv == NULL && mn.state != MN_DISCONNECTED) {
		find_agent(STATE_INIT);
	}
}

/**
 * check_interfaces:
 * @iface: array of interfaces
 * @iface_num: number of interfaces in @iface
 * 
 * Check whether new interfaces can be detected or if an interface 
 * has been removed
 *
 * Returns: 0 if successfully executed, 1 on error
 */
int check_interfaces(struct interface_data *iface, int iface_num)
{
	int i;
	struct idxmap *idxmap, *idx;

	idxmap = dyn_ip_get_interface_map();
	if (idxmap == NULL) {
		DEBUG(DEBUG_INFO, "check_interfaces: interface map could "
		      "not be created\n");
		return 1;
	}

	/* remove the ignored interfaces */
	/* FIX: add list of interfaces to be ignored to dynmnd.conf and try to
	 * eliminate 'virtual' interfaces like tunnels etc. */
	idx = idxmap;
	while (idx != NULL) {
		if (iface_is_cr_tnl(idx->name)		||
			check_ignore_iface(idx->name) 	||
		    (dyn_ip_get_ifflags(idx->name) & IFF_UP) == 0)
			idx->index = -1;
		idx = idx->next;
	}

	/* check removed interfaces */
	for (i = 0; i < iface_num; i++) {
		idx = idxmap;
		while (idx != NULL) {
			if (iface[i].index == idx->index)
				break;
			idx = idx->next;
		}
		
		if (iface[i].s > -1) {
			if (idx != NULL) {
				/* mark the interface used */
				idx->index = -1;
			} else {			
				DEBUG(DEBUG_INFO, "check_interfaces: "
				      "removed interface\n");
				/* this interface has been removed */
				modify_removed_interface(&iface[i]);
			}
		}
	}
	
	/* check new interfaces */
	idx = idxmap;
	while (idx != NULL) {
		if (idx->index == -1) {
			idx = idx->next;
			continue;
		}

		/* find a free slot */
		i = 0;
		while (i < iface_num && iface[i].s > -1) i++;
		if (i >= iface_num) {
			DEBUG(DEBUG_INFO, "Not enough space for all "
			      "the interfaces\n");
			break;
		}

		/* call handlers for INTERFACE_INIT event */
		DEBUG(DEBUG_INFO, "check_interfaces: new interface\n");
		modify_new_interface(&iface[i], idx);
		idx = idx->next;
	}

	dyn_ip_free_interface_map(idxmap);

	return 0;
}
/**
 * create_registration_socket:
 *
 * Create registration socket and bind it to specified port.
 *
 * Returns: created socket or -1 on error. 
 */
int create_registration_socket(void)
{
	int s;
	int one = 1;
	unsigned char ttl;
#ifdef BIND_UDP_SOCKET
	struct sockaddr_in addr;
#endif

	if ((s = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		LOG2(LOG_ERR, "create_reg_socket - socket: %s\n",
		     strerror(errno));
		return -1;
	}
	/* Set IP packet info */
	if (setsockopt(s, SOL_IP, IP_PKTINFO, &one, sizeof(one)) < 0)
	{
    	perror("setsockopt(IP_PKTINFO): ");
	}
#ifdef BIND_UDP_SOCKET
	memset(&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = config.bind_addr.s_addr;
	addr.sin_port = config.bind_port;
	DEBUG(DEBUG_INFO, "binding UDP socket to %s:%i\n",
	      inet_ntoa(addr.sin_addr), ntohs(addr.sin_port));
	if (bind(s, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
		LOG2(LOG_ERR, "create_reg_socket - bind: %s\n",
		     strerror(errno));
		return -1;
	}
#endif

	if (config.socket_priority > -1 &&
	    setsockopt(s, SOL_SOCKET, SO_PRIORITY,
		       (char *)&config.socket_priority,
		       sizeof(config.socket_priority)) < 0) {
		LOG2(LOG_ERR,
		     "create_reg_socket - setsockopt SO_PRIORITY: %s\n",
		     strerror(errno));
		/* continue without SO_PRIORITY set */
	}

	/* send registration request with TTL 255 [RFC 2344, Chap. 3.2] */
	ttl = 255;
	if (setsockopt(s, IPPROTO_IP, IP_TTL, &ttl, sizeof(ttl)) < 0) {
		LOG2(LOG_ERR, "setsockopt: IP_TTL failed: %s\n",
		     strerror(errno));
		/* continue with the default setting anyway */
#ifdef DYN_TARGET_WINDOWS
		if (dyn_ip_set_ttl(255) < 0) {
			DEBUG(DEBUG_INFO, "Also dyn_ip_set_ttl() failed\n");
		}
#endif /* DYN_TARGET_WINDOWS */
	}

	return s;
}


static void show_usage(void)
{
	printf("      --wireless_policy <integer>\n"
	       "\t\tset wireless handoff policy (1=on, 0=off):\n"
	       "\t\t  bit 0: rely on MN's agent adv. expiration\n"
	       "\t\t  bit 1: use shorter agent adv. expiration interval\n"
	       "\t\t  bit 2: use FA with newest agent advertisement\n"
	       "\t\t  bit 3: eager switching\n"
	       "\t\t  bit 4: use newest agent advertisement\n"
	       "      --no-wireless\tdisable wireless extensions\n"
	       "      --disconnect\tstart in disconnected state\n");
}

int mn_parse_command_line(int argc, char *argv[])
{
	int c, oindex = 0;

	static struct option long_options[] = {
		{"help", no_argument, NULL, 'h'},
		{"version", no_argument, NULL, 'v'},
		{"debug", no_argument, NULL, 0},
		{"fg", no_argument, NULL, 0},
		{"config", required_argument, NULL, 'c'},
		{"traces", no_argument, NULL, 't'},	
		/* MN specific arguments */
		{"no-wireless", no_argument, NULL, 'n'},
		{"wireless_policy", required_argument, NULL, 'p'},
		{"disconnect", no_argument, NULL, 'd'},
		/* end of arguments */
		{0, 0, 0, 0}
	};

	mn.opt_connect = 1;
	mn.policy_bits = 0;

	/* parse mn specific options */
	wireless_extensions = 1;
	while ((c = getopt_long(argc, argv, "+hv", long_options, &oindex)) !=
	       EOF) {
		switch (c) {
		case 't':
			printf("--traces: Print traces.\n");
			mn_traces = 1;
			break;
		case 'n':
			wireless_extensions = 0;
			break;
		case 'p':
			mn.policy_bits = atoi(optarg);
			/* 0: no timestamps */
			/* 1: with timestamps */
			DEBUG(DEBUG_INFO, "Option: %s"
			      " 0x%02x (wireless policy)\n",
			      long_options[oindex].name,
			      mn.policy_bits);
			break;
		case 'h':
			/* show MN specific parameters */
			show_usage();
			exit(1);
			break;

		case 'd':
			mn.opt_connect = 0;
			break;

		case '?':
			printf("Command line parsing failed - aborting\n");
			exit(1);

		case 'v':
		case 'c':
		case 0:
			break;

		default:
			printf("getopt returned character code 0%o\n", c);
			exit(1);
		}
	}

	return 0;
}


#ifdef DYN_TARGET_WINDOWS
/* FIX: currently using local UDP socket on Windows because there is no support
 * for UNIX domain socket.. maybe something else would be better for IPC on
 * Windows platform(?) */
static int windows_api_socket(int admin)
{
	int s;
	struct sockaddr_in addr;
	int flags;

	s = socket(AF_INET, SOCK_DGRAM, 0);
	if (s < 0) {
		DEBUG(DEBUG_INFO, "API socket opening failed: %s\n",
		      strerror(errno));
		return -1;
	}

	flags = fcntl(s, F_GETFL, 0);
	if (flags == -1 || fcntl(s, F_SETFL, flags | O_NDELAY) == -1) {
		DEBUG(DEBUG_INFO, "API rw fcntl F_SETFL failed: %s\n",
		      strerror(errno));
		close(s);
		return -1;
	}

	memset(&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = htonl((127 << 24) | 1);
	addr.sin_port = htons(admin ? API_UDP_PORT_RW : API_UDP_PORT_RO);
	if (bind(s, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
		LOG2(LOG_ERR, "API socket - bind: %s\n", strerror(errno));
		close(s);
		return -1;
	}

	return s;
}
#endif

/* Initialization routines. Setup sockets and variables.
 * Returns 1 on success else 0. */
int mn_init(void)
{
	int i;
	
	/* Load configuration */
	DEBUG(DEBUG_INFO, "Load config\n");
	if (!load_mn(&config, program_name,
		     opt_config == NULL ? MN_GLOBAL_CONF_FILE : opt_config)) {
		LOG2(LOG_ERR, "init: unable to load configuration\n");
		return 0;
	}
	mn.use_auth_alg = config.auth_alg;
	mn.agentadv = hashtable_init(256);
	if (mn.agentadv == NULL) {
		LOG2(LOG_ERR, "init: unable to init agentadv hashtable\n");
		return 0;
	}
	mn.last_req_FA_NAI = (struct fa_nai_ext *) malloc(MAX_FA_NAI_LEN);
	if (mn.last_req_FA_NAI == NULL) {
		LOG2(LOG_ERR, "init: not enough memory\n");
		return 0;
	}
	memset(mn.last_req_FA_NAI, 0, MAX_FA_NAI_LEN);

	if (!config.enable_fa_decapsulation)
		check_kernel_support(CHECK_KERNEL_IPIP);
	check_kernel_support(CHECK_KERNEL_NETLINK);

#ifdef DYN_TARGET_LINUX
	if (getuid() || geteuid()) {
		fprintf(stderr, "This program must be run by root.\n");
		return 0;
	}
#endif
#ifdef DYN_TARGET_WINDOWS
	/* FIX: on Windows NT and Windows 2000 check for Administrator
	 * privileges */
#endif

	if ((mn.registration_socket = create_registration_socket()) < 0)
		return 0;
	
	mn_handlers_init();
	mn.dev_info_addr.sun_family = AF_LOCAL;
	/* initialize default socket info path */
	dynamics_strlcpy(mn.dev_info_addr.sun_path, DEFAULT_DEVICE_INFO_PATH,
			 sizeof(mn.dev_info_addr.sun_path));

#ifdef WITH_WIRELESS
	/* If wireless extensions enabled, monitor will register it's own
	   handlers also */
	DEBUG(DEBUG_INFO, "REGISTER MODULE(S)\n");
	if (wireless_extensions)
		MONITOR_register_module();
#endif

	/* Check interfaces */
	for (i = 0; i < MAX_INTERFACES; i++) {
		mn.iface[i].s = -1;
		mn.iface[i].s_adv = -1;
		mn.iface[i].handlers_off = 1;
	}

#ifdef INCLUDE_IPAY
	/* register Ipay handlers after monitor */
	ipay_init();
#endif

	if (config.home_net_addr_plen < 0 ||
	    dyn_ip_route_get(config.home_net_addr, mn.start_default_device,
			     IFNAMSIZ) != 0)
		mn.start_default_device[0] = '\0';
	else
		DEBUG(DEBUG_INFO, "Initial default device: %s\n",
		      mn.start_default_device);

#ifdef DYN_TARGET_LINUX
	if ((mn.api_ro_socket =
	     api_open_socket(config.mn_api_read_socket_path,
			     config.mn_api_read_socket_group,
			     config.mn_api_read_socket_owner,
			     config.mn_api_read_socket_permissions)) < 0) {
		LOG2(LOG_ERR, "init - open_api_ro_socket: %s\n",
		     strerror(errno));
		return 0;
	}

	if ((mn.api_rw_socket =
	     api_open_socket(config.mn_api_admin_socket_path,
			     config.mn_api_admin_socket_group,
			     config.mn_api_admin_socket_owner,
			     config.mn_api_admin_socket_permissions)) < 0) {
		LOG2(LOG_ERR, "init - open api_rw_socket: %s\n",
		     strerror(errno));
		return 0;
	}
#endif
#ifdef DYN_TARGET_WINDOWS
	mn.api_ro_socket = windows_api_socket(0);
	mn.api_rw_socket = windows_api_socket(1);
#endif

#ifdef INCLUDE_IPAY
	mn.ipay_sock = socket(AF_INET, SOCK_DGRAM, 0);
	if (mn.ipay_sock < 0) {
		LOG2(LOG_ERR, "init - open ipay_sock: %s\n", strerror(errno));
		return 0;
	}
	if (config.ipay_mn_port > 0) {
		struct sockaddr_in addr;
		memset(&addr, 0, sizeof(addr));
		addr.sin_family = AF_INET;
		addr.sin_addr = config.ipay_mn_addr;
		addr.sin_port = htons(config.ipay_mn_port);
		if (bind(mn.ipay_sock, (struct sockaddr *) &addr,
		    sizeof(addr)) < 0) {
			LOG2(LOG_ERR, "init - bind ipay_sock: %s\n",
			     strerror(errno));
			return 0;
		}
	}
	mn.ipay_sock_fa = socket(AF_INET, SOCK_DGRAM, 0);
	if (mn.ipay_sock_fa < 0) {
		LOG2(LOG_ERR, "init - open ipay_sock_fa: %s\n",
		     strerror(errno));
		return 0;
	}
	if (config.ipay_fa_port > 0) {
		struct sockaddr_in addr;
		memset(&addr, 0, sizeof(addr));
		addr.sin_family = AF_INET;
		addr.sin_addr.s_addr = INADDR_ANY;
		addr.sin_port = htons(config.ipay_fa_port);
		if (bind(mn.ipay_sock_fa, (struct sockaddr *) &addr,
		    sizeof(addr)) < 0) {
			LOG2(LOG_ERR, "init - bind ipay_sock_fa: %s\n",
			     strerror(errno));
			return 0;
		}
	}
#endif

	mn.rtnetlink_socket = dyn_ip_monitor_open();
	if (mn.rtnetlink_socket < 0)
		DEBUG(DEBUG_INFO, "init - rtnetlink socket opening failed\n");

	/* jpaaso - This does not work if the iface's address != config.mn_home_ip_addr!
	 * The value is updated before jumping into the main loop */
	/* set own ip-addr */
	mn.local_addr = config.mn_home_ip_addr;

	init_tunneling();

	/* init variables */
	mn.registration_id[0] = 0;
	mn.registration_id[1] = 0;
	mn.last_ha_nonce = 0;
	mn.clock_correction = 0;
	mn.state = MN_DISCONNECTED;
	mn.fa_addr.s_addr = 0;
	mn.current_adv = NULL;
	mn.force_fa_addr.s_addr = 0;
	mn.session_key = NULL;
	mn.session_key_len = 0;
	mn.tunnel_up = 0;
	mn.tunnel_mode = API_TUNNEL_NONE;
	mn.ha_error_count = 0;
	memset(&mn.cur_route_info, 0, sizeof(mn.cur_route_info));
	mn.cur_route_info.net.s_addr = config.home_net_addr.s_addr;
	mn.cur_route_info.prefix_len = config.home_net_addr_plen;
	timerclear(&mn.last_reg_send_time);
	for (i = 0; i < TIMER_COUNT; i++) timerclear(&timers[i]);
	
	if (config.wlan_ap_poll_interval > -1)
		gettimeofday(&timers[TIMER_WLAN_AP_POLL], NULL);
	if (config.solicitation_interval > -1)
		gettimeofday(&timers[TIMER_SOLICITATION], NULL);

        openlog(SYSLOG_IDENT, SYSLOG_OPTIONS, config.syslog_facility);
	
	/* jpaaso - saving default route disabled
	dyn_ip_route_save_default(config.home_net_gateway.s_addr != 0 ?
		&config.home_net_gateway : NULL);
	*/
#ifdef MN_LOCUPD_PROFILER
	mn.profile = fopen("mn.prof.log", "a");
	if (mn.profile == NULL) {
		fprintf(stderr, "Cannot open mn.prof.log\n");
		return 0;
	}
	gettimeofday(&mn.last_api, NULL);
	fprintf(mn.profile, "START %u.%06u\n",
		(unsigned int) mn.last_api.tv_sec,
		(unsigned int) mn.last_api.tv_usec);
#endif
	
	/* jpaaso - TODO: this does not seem to work that well */
    /* Setup signals*/
    signal(SIGTERM, clean_up);
    signal(SIGINT, clean_up);
    signal(SIGHUP, SIG_IGN);
    signal(SIGPIPE, SIG_IGN);

	if (!opt_foreground && dynamics_fork_daemon() == -1)
		clean_up(0);
	dynamics_write_pid_file(MN_PID_FILE);

    DEBUG(DEBUG_INFO, " MN initialized to %sconnected state,\n"
	      "  SPI: %d, HA: %s,\n", (mn.opt_connect ? "" : "dis"),
	      config.spi, inet_ntoa(config.ha_ip_addr));
	DEBUG(DEBUG_INFO, "  HomeAddr: %s",
	      inet_ntoa(config.mn_home_ip_addr));
	DEBUG(DEBUG_INFO, ", CurrentAddr: %s\n", inet_ntoa(mn.local_addr));

	syslog(LOG_INFO, "Started.");

	check_interfaces(mn.iface, MAX_INTERFACES);

#ifdef DYN_TARGET_WINDOWS
	mn.pcap_capturer = init_pcap_for_advs();
#else
	mn.pcap_capturer = -1;
#endif /* DYN_TARGET_WINDOWS */

	return 1;
}

/* Perform clean up: close sockets and log. */
void clean_up(int sig)
{
	int i;

	print_trace(mn_traces, "cleaning up..\n");
	
	/* Deregister to CRs */
	clean_ro_bindings();

	/* try to disconnect if connected */
	if (mn.state == MN_CONNECTED) {
		send_registration(REG_DISC);
		DEBUG(DEBUG_MESSAGES, "Deregistration message sent\n");
	}
	
	if (mn.pcap_capturer > -1) {
		DEBUG(DEBUG_INFO, "Terminating pcap capturer\n");
		kill(mn.pcap_capturer, SIGTERM);
		usleep(100000);
		kill(mn.pcap_capturer, SIGKILL);
	}

	if (mn.agentadv != NULL) {
		hashtable_iterator(mn.agentadv, clean_agentadv, NULL);
		hashtable_destroy(mn.agentadv);
	}

	/* close sockets */
	close(mn.registration_socket);
	close(mn.hoa_socket);
	close(mn.hoa_listener_socket);
	
	for (i = 0; i < NW_IFACE_ARRAY_LEN; i++) {
		if (usable_nw_ifaces[i].listener_socket > 0) 
			close(usable_nw_ifaces[i].listener_socket);
	}
	for (i = 0; i < ro_entry_count; i++) {
		close(ro_cache[i].registration_socket);
		close(ro_cache[i].coti_socket);
	}
	
	for (i = 0; i < MAX_INTERFACES; i++) {
		if (mn.iface[i].s > -1)
			close(mn.iface[i].s);
		if (mn.iface[i].s_adv > -1)
			close(mn.iface[i].s_adv);
	}
	if (mn.api_ro_socket > -1)
		close(mn.api_ro_socket);
	if (mn.api_rw_socket > -1)
		close(mn.api_rw_socket);
#ifdef INCLUDE_IPAY
	close(mn.ipay_sock);
	ipay_cleanup();
#endif
	if (mn.rtnetlink_socket > -1)
		close(mn.rtnetlink_socket);

#ifdef WITH_WIRELESS
	/* Remove iwspy entries */
	MONITOR_unregister_module();
#endif

	/* stop to any waiting api call */
	reply_waiting_api(API_FAILED, NULL, 0);
	/* close tunneling device */
	close_tunneling();
	mn_remove_dynamic_home_addr();

	/* free session key */
	if (mn.session_key != NULL)
		free(mn.session_key);

	if (mn.last_req_FA_NAI != NULL)
		free(mn.last_req_FA_NAI);

	/* unlink api domain sockets */
	unlink(config.mn_api_read_socket_path);
	unlink(config.mn_api_admin_socket_path);

	unlink(MN_PID_FILE);

	/* unregister all handlers */
	/* FA_ADV_RECEIVE */
	i = handler_unregister_all(FA_ADV_RECEIVE);
	DEBUG(DEBUG_HANDLERS, "clean_up: %d \"%s\" handler(s) "
	      "removed\n", i, event_type_str(FA_ADV_RECEIVE));
	/* FA_ADV_EXPIRE  */
	i = handler_unregister_all(FA_ADV_EXPIRE);
	DEBUG(DEBUG_HANDLERS, "clean_up: %d \"%s\" handler(s) "
	      "removed\n", i, event_type_str(FA_ADV_EXPIRE));
	/* FA_GET */
	i = handler_unregister_all(FA_GET);
	DEBUG(DEBUG_HANDLERS, "clean_up: %d \"%s\" handler(s) "
	      "removed\n", i, event_type_str(FA_GET));
	/* INTERFACE_INIT */
	i = handler_unregister_all(INTERFACE_INIT);
	DEBUG(DEBUG_HANDLERS, "clean_up: %d \"%s\" handler(s) "
	      "removed\n", i, event_type_str(INTERFACE_INIT));
	/* INTERFACE_DOWN */
	i = handler_unregister_all(INTERFACE_DOWN);
	DEBUG(DEBUG_HANDLERS, "clean_up: %d \"%s\" handler(s) "
	      "removed\n", i, event_type_str(INTERFACE_DOWN));

	syslog(LOG_INFO, "stopped");
	closelog();

#ifdef MN_LOCUPD_PROFILER
	fclose(mn.profile);
#endif

	DEBUG(DEBUG_INFO, "cleaned up\n");

	exit(0);
}


void add_fa_host_route(char *dev, struct hashtable *adv_hash, 
		       int ifindex, struct in_addr addr)
{
	struct agentadv_data *adv;
	adv = adv_fetch(adv_hash, &addr, ifindex);

	if (dyn_ip_route_replace(addr, dev) != 0) {
		if (adv != NULL && adv->routeentry)
			DEBUG(DEBUG_INFO, "Route entry already added for the "
			      "FA (%s=>%s)\n", inet_ntoa(addr), dev);
		else
			DEBUG(DEBUG_INFO, "Could not add a route to the FA "
			      "(%s=>%s), but trying to register anyway\n",
			      inet_ntoa(addr), dev);
	} else {
		if (adv != NULL)
			adv->routeentry = 1;
		else
			DEBUG(DEBUG_INFO, "No agentadv for FA %s\n",
			      inet_ntoa(addr));
	}
}


static int clean_routes(struct node *node, void *data)
{
	struct agentadv_data *adv, *current;
	adv = (struct agentadv_data *) node;
	current = (struct agentadv_data *) data;

	if (current == adv) return TRUE; /* do not remove current FA's entry */
	if (!adv->routeentry) return TRUE;

	DEBUG(DEBUG_INFO, "Removing FA route (%s,%s)\n",
	      inet_ntoa(adv->addr), adv->ifname);
	if (dyn_ip_route_del(adv->addr, adv->ifname) != 0)
		LOG2(LOG_ERR, "FA route removing failed (addr=%s, dev=%s)\n",
		     inet_ntoa(adv->addr), adv->ifname);

	adv->routeentry = 0;
	return TRUE;
}


void remove_fa_host_routes(int all)
{
	DEBUG(DEBUG_INFO, "remove_fa_host_routes(%i)\n", all);
	hashtable_iterator(mn.agentadv, clean_routes,
			   all ? NULL: mn.current_adv);
}



int is_coloc_addr_foreign(void)
{
	char dev[IFNAMSIZ];
	if (dyn_ip_route_get(config.mn_home_ip_addr, dev, IFNAMSIZ) != 0 ||
	    strcmp(dev, "lo") != 0) 
	{
		DEBUG(DEBUG_INFO, "[mn_util.c] is_coloc_addr_foreign() - Yes\n");
		return 1;
	}
	DEBUG(DEBUG_INFO, "[mn_util.c] is_coloc_addr_foreign() - No\n");
	return 0;
}


int device_up(int ifindex)
{
	int i;
	for (i = 0; i < MAX_INTERFACES; i++) {
		if (mn.iface[i].s > -1 &&
		    mn.iface[i].index == ifindex)
			return 1;
	}
	return 0;
}


char *event_type_str(int event_type)
{
	switch (event_type) {
	case FA_ADV_RECEIVE:
		return "FA Advertisement receive";
	case FA_ADV_EXPIRE:
		return "FA Advertisement expire";
	case FA_GET:
		return "Get best FA";
	case INTERFACE_INIT:
		return "Initialize interface";
	case INTERFACE_DOWN:
		return "Interface down";
	default:
		return "Unknown event type!";
	}
}



/**
 * mn_remove_dynamic_home_addr:
 *
 * Remove dynamic home address from interface if it is maintained by
 * Dynamics MN.
 *
 * Returns: 1 on success (address removed), 0 if no dynamic home address, or
 * -1 on failure
 */
int mn_remove_dynamic_home_addr(void)
{
	if (mn.dynamic_home_addr.s_addr == 0)
		return 0;

	if (dyn_ip_addr_del(mn.dynamic_home_addr_dev, mn.dynamic_home_addr) ==
	    0) {
		DEBUG(DEBUG_INFO, "Dynamic home address '%s' removed from "
		      "interface '%s'\n", inet_ntoa(mn.dynamic_home_addr),
		      mn.dynamic_home_addr_dev);
		if (config.mn_home_ip_addr.s_addr ==
		    mn.dynamic_home_addr.s_addr) {
			DEBUG(DEBUG_INFO, "\treset home address to 0 (use "
			      "discovery)\n");
			config.mn_home_ip_addr.s_addr =
				mn.local_addr.s_addr = 0;
			if (config.ha_ip_addr_orig.s_addr == 0) {
				DEBUG(DEBUG_INFO, "\treset home agent address "
				      "to 0 (use discovery)\n");
				config.ha_ip_addr.s_addr = 0;
			}
		} else {
			DEBUG(DEBUG_INFO, "\tDEBUG: config: %s, ",
			      inet_ntoa(config.mn_home_ip_addr));
			DEBUG(DEBUG_INFO, "obtained: %s\n",
			      inet_ntoa(mn.dynamic_home_addr));
		}

		mn.dynamic_home_addr.s_addr = 0;
		return 1;
	}

	DEBUG(DEBUG_INFO, "Could not remove dynamic home address '%s' "
	      "from interface '%s'\n", inet_ntoa(mn.dynamic_home_addr),
	      mn.dynamic_home_addr_dev);
	return -1;
}


int monitor_check_policy(int bit)
{
	return POLICY(bit);
}


int copy_str(char *buffer, int len, int *curr, char *str, char *attr)
{
	int l;

	l = strlen(str) + strlen(attr) + 1;
	if (*curr+l > len) {
		DEBUG(DEBUG_INFO, "copy_str: Not enough room (%d/%d)\n", 
		      l, len);
		return 1;
	}
	sprintf(buffer+*curr, "%s%s\n", str, attr);
	*curr += l;

	return 0;
}
