/* $Id: auth.c,v 1.20 2001/08/03 22:34:09 jm Exp $
 * Message Authentication Routines
 *
 * Dynamic hierarchial IP tunnel
 * Copyright (C) 1998-2001, Dynamics group
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation. See README and COPYING for
 * more details.
 */

#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "md5_mac.h"
#include "sha1.h"
#include "debug.h"
#include "util.h"
#include "auth.h"

#define DEBUG_FLAG 'U'
/* #define AUTH_EXTRA_DEBUG */


/* RFC 2002, sec. 3.5.1.
 * MAC protects UDP payload (reg. req or reg. reply),
 * all prior extension, and Type and Length of this auth. extension
 *
 * draft-ietf-mobileip-v2-00.txt adds SPI field of the extension to the
 * protected area and this is used as a default in our implementation.
 * There is an automatic fallback to use the RFC 2002 method to provide
 * compatibility with other implementations.
 *
 * Note: data pointer should be to the beginning of the registration
 * request or reply and auth pointer must be in the same message buffer
 * (i.e. just after the data to be protected)
 */


#ifdef AUTH_EXTRA_DEBUG

static void show_hex(unsigned char *data, int len)
{
	int i;
	for (i = 0; i < len; i++)
		DEBUG(DEBUG_FLAG, "%02x", data[i]);
}

#endif


typedef void (*MACfunc)(unsigned char *key, unsigned int key_len,
			unsigned char *data, unsigned int data_len,
			unsigned char *mac);

static MACfunc auth_get_mac_data(int alg, int *mac_len)
{
	if (alg == AUTH_ALG_MD5 || alg == AUTH_ALG_MD5_RFC2002 ||
	    alg == AUTH_ALG_MD5_PREFIX_SUFFIX) {
		*mac_len = MD5_MAC_LEN;
		return md5_mac;
	} else if (alg == AUTH_ALG_HMAC_MD5) {
		*mac_len = MD5_MAC_LEN;
		return hmac_md5;
	} else if (alg == AUTH_ALG_SHA1) {
		*mac_len = SHA1_MAC_LEN;
		return sha1_mac;
	} else if (alg == AUTH_ALG_HMAC_SHA1) {
		*mac_len = SHA1_MAC_LEN;
		return hmac_sha1;
	}

	return NULL;
}


/* Add an authentication extension to a message and return the length of the
 * added authentication extension.
 * Parameter spi must be in network byte order.
 */
int auth_add(int alg, unsigned char *key, unsigned int keylen,
	     unsigned char *data, struct msg_auth *auth,
	     __u8 auth_type, __u32 spi)
{
	int len;
	int mac_len;
	MACfunc mac_func;

	auth->type = auth_type;
	auth->spi = spi;

	mac_func = auth_get_mac_data(alg, &mac_len);
	if (mac_func != NULL) {
		auth->length = mac_len + SPI_LEN;
		len = ((unsigned char *)auth) - data +
			(alg == AUTH_ALG_MD5_RFC2002 ? 2 : 6);
		mac_func(key, keylen, data, len, MSG_AUTH_DATA(auth));
	} else {
		DEBUG(DEBUG_FLAG, "auth_add: unknown algorithm %i\n", alg);
		auth->length = SPI_LEN;
		mac_len = 0;
	}

#ifdef AUTH_EXTRA_DEBUG
	DEBUG(DEBUG_FLAG, "auth_add: len=%i, mac=", len);
	show_hex(MSG_AUTH_DATA(auth), mac_len);
	DEBUG(DEBUG_FLAG, "\n");
#endif

	return GET_AUTH_EXT_LEN(auth);
}


/* returns zero on failure and nonzero on successful authentication check */
int auth_check(int alg, unsigned char *key, unsigned int keylen,
	       unsigned char *data, struct msg_auth *auth)
{
	int len, ret;
	unsigned char *mac;
	int mac_len;
	MACfunc mac_func;

	mac_func = auth_get_mac_data(alg, &mac_len);
	if (mac_func == NULL) {
		DEBUG(DEBUG_FLAG, "auth_check: unknown algorithm %i\n", alg);
		return 0;
	}

	mac = (unsigned char *) malloc(mac_len);
	if (mac == NULL) {
		DEBUG(DEBUG_FLAG, "auth_check: malloc(%i) failed\n", mac_len);
		return 0;
	}

	if (auth->length != mac_len + SPI_LEN) {
		DEBUG(DEBUG_FLAG, "[other/auth.c] auth_check: len=%i != %i\n",
		      auth->length, mac_len + SPI_LEN);
		return 0;
	}

	len = ((unsigned char *)auth) - data +
		(alg == AUTH_ALG_MD5_RFC2002 ? 2 : 6);
	mac_func(key, keylen, data, len, mac);
#ifdef AUTH_EXTRA_DEBUG
	DEBUG(DEBUG_FLAG, "auth_check: len=%i, recvMAC=", len);
	show_hex(MSG_AUTH_DATA(auth), mac_len);
	DEBUG(DEBUG_FLAG, ", expMAC=");
	show_hex(mac, mac_len);
	DEBUG(DEBUG_FLAG, "\n");
#endif

	ret = memcmp(mac, MSG_AUTH_DATA(auth), mac_len) == 0;
	free(mac);

	return ret;
}


int auth_add_vendor(int alg, unsigned char *key, unsigned int keylen,
		    unsigned char *data, struct vendor_msg_auth *auth,
		    __u16 auth_type, __u32 spi)
{
	int len;
	int mac_len;
	MACfunc mac_func;

	auth->type = VENDOR_EXT_TYPE2;
	auth->reserved = 0;
	auth->vendor_id = htonl(VENDOR_ID_DYNAMICS);
	auth->sub_type = htons(auth_type);
	auth->spi = spi;

	mac_func = auth_get_mac_data(alg, &mac_len);
	if (mac_func != NULL) {
		auth->length = mac_len + MIN_VENDOR_AUTH_LEN;
		len = ((unsigned char *)auth) - data +
			sizeof(struct vendor_msg_auth);
		mac_func(key, keylen, data, len, MSG_VENDOR_AUTH_DATA(auth));
	} else {
		DEBUG(DEBUG_FLAG, "auth_add_vendor: unknown algorithm %i\n",
		      alg);
		auth->length = MIN_VENDOR_AUTH_LEN;
		mac_len = 0;
	}

#ifdef AUTH_EXTRA_DEBUG
	DEBUG(DEBUG_FLAG, "auth_add_vendor: len=%i, mac=", len);
	show_hex(MSG_VENDOR_AUTH_DATA(auth), mac_len);
	DEBUG(DEBUG_FLAG, "\n");
#endif

	return GET_VENDOR_AUTH_EXT_LEN(auth);
}


int auth_check_vendor(int alg, unsigned char *key, unsigned int keylen,
		      unsigned char *data, struct vendor_msg_auth *auth)
{
	int len, ret;
	unsigned char *mac;
	int mac_len;
	MACfunc mac_func;

	mac_func = auth_get_mac_data(alg, &mac_len);
	if (mac_func == NULL) {
		DEBUG(DEBUG_FLAG, "auth_check_vendor: unknown algorithm %i\n",
		      alg);
		return 0;
	}

	mac = (unsigned char *) malloc(mac_len);
	if (mac == NULL) {
		DEBUG(DEBUG_FLAG, "auth_check_vendor: malloc(%i) failed\n",
		      mac_len);
		return 0;
	}

	if (auth->length != mac_len + MIN_VENDOR_AUTH_LEN) {
		DEBUG(DEBUG_FLAG, "auth_check_vendor: len=%i != %i\n",
		      auth->length, mac_len + MIN_VENDOR_AUTH_LEN);
		return 0;
	}

	len = ((unsigned char *)auth) - data +
		sizeof(struct vendor_msg_auth);
	mac_func(key, keylen, data, len, mac);
#ifdef AUTH_EXTRA_DEBUG
	DEBUG(DEBUG_FLAG, "auth_check_vendor: len=%i, recvMAC=", len);
	show_hex(MSG_VENDOR_AUTH_DATA(auth), mac_len);
	DEBUG(DEBUG_FLAG, ", expMAC=");
	show_hex(mac, mac_len);
	DEBUG(DEBUG_FLAG, "\n");
#endif
	ret = memcmp(mac, MSG_VENDOR_AUTH_DATA(auth), mac_len) == 0;
	free(mac);

	return ret;
}


/* Add a generalized authentication extension into the message and return the
 * length of the added authentication extension.
 * Parameter spi must be in network byte order.
 */
int auth_add_gen(int alg, unsigned char *key, unsigned int keylen,
		 unsigned char *data, unsigned char *challenge,
		 int challenge_len, struct generalized_auth_ext *auth,
		 __u8 auth_subtype, __u32 spi)
{
	int len;
	int mac_len;
	MACfunc mac_func;

	auth->type = GENERALIZED_AUTH_EXT;
	auth->subtype = auth_subtype;
	auth->spi = spi;

	if (alg == AUTH_ALG_RADIUS) {
		unsigned char tmp[MD5_MAC_LEN];
		MD5_CTX context;

		auth->length = htons(MD5_MAC_LEN + SPI_LEN);

		MD5Init(&context);
		len = ((unsigned char *) auth) - data + 8;
		MD5Update(&context, data, len);
		MD5Final(tmp, &context);

		if (challenge_len > 0)
			MD5Update(&context, challenge, 1);
		MD5Update(&context, key, keylen);
		MD5Update(&context, tmp, MD5_MAC_LEN);
		if (challenge_len > 1)
			MD5Update(&context, challenge + 1, challenge_len - 1);
		MD5Final(MSG_GEN_AUTH_DATA(auth), &context);
	} else {
		mac_func = auth_get_mac_data(alg, &mac_len);
		if (mac_func != NULL) {
			auth->length = htons(mac_len + SPI_LEN);
			len = ((unsigned char *) auth) - data + 8;
			mac_func(key, keylen, data, len,
				 MSG_GEN_AUTH_DATA(auth));
		} else {
			DEBUG(DEBUG_FLAG, "auth_add_gen: unknown algorithm %i"
			      "\n", alg);
			auth->length = htons(SPI_LEN);
		}
	}

	return GET_GEN_AUTH_EXT_LEN(auth);
}


int auth_check_gen(int alg, unsigned char *key, unsigned int keylen,
		   unsigned char *data, struct generalized_auth_ext *auth)
{
	int len, ret;
	unsigned char *mac;
	int mac_len;
	MACfunc mac_func;

	/* FIX: add support for AUTH_ALG_RADIUS as a special case like with
	 * auth_add_gen() above (?) */

	mac_func = auth_get_mac_data(alg, &mac_len);
	if (mac_func == NULL) {
		DEBUG(DEBUG_FLAG, "auth_check_gen: unknown algorithm %i\n",
		      alg);
		return 0;
	}

	mac = (unsigned char *) malloc(mac_len);
	if (mac == NULL) {
		DEBUG(DEBUG_FLAG, "auth_check_gen: malloc(%i) failed\n",
		      mac_len);
		return 0;
	}

	if (ntohs(auth->length) != mac_len + SPI_LEN) {
		DEBUG(DEBUG_FLAG, "auth_check_gen: len=%i != %i\n",
		      ntohs(auth->length), mac_len + SPI_LEN);
		return 0;
	}

	len = ((unsigned char *)auth) - data +
		sizeof(struct generalized_auth_ext);
	mac_func(key, keylen, data, len, mac);
#ifdef AUTH_EXTRA_DEBUG
	DEBUG(DEBUG_FLAG, "auth_check_gen: len=%i, recvMAC=", len);
	show_hex(MSG_GEN_AUTH_DATA(auth), mac_len);
	DEBUG(DEBUG_FLAG, ", expMAC=");
	show_hex(mac, mac_len);
	DEBUG(DEBUG_FLAG, "\n");
#endif
	ret = memcmp(mac, MSG_GEN_AUTH_DATA(auth), mac_len) == 0;
	free(mac);

	return ret;
}


/* Encrypts the session key with given algorithm and shared secret */
int auth_encrypt(int alg, unsigned char *ssec, unsigned int sseclen,
		 unsigned char *key, struct msg_key *keyrep,
		 struct reg_rep *data, __u8 keytype, __u32 spi)
{
	unsigned char *t;
	int i;
	int mac_len;
	MACfunc mac_func;

	mac_func = auth_get_mac_data(alg, &mac_len);
	if (mac_func == NULL) {
		DEBUG(DEBUG_FLAG, "auth_encrypt: unknown algorithm %i\n", alg);
		return SPI_LEN;
	}

	init_key_extension(keyrep, keytype, spi, mac_len);
	t = MSG_KEY_DATA(keyrep);
	mac_func(ssec, sseclen, (unsigned char *) data, sizeof(struct reg_rep),
		 t);
	for (i = 0; i < mac_len; i++)
		t[i] ^= key[i];
	return GET_KEY_EXT_LEN(keyrep);
}


/* Decrypts keyrep extension with given algorithm and shared secret
 * Returns:
 *    Pointer to the decrypted session (caller must release the memory) on
 *    success or NULL on failure.
 *    sk_len is set to the length of the session key or on failure to an
 *    error code:
 *       -1: could not allocate enough memory
 *       -2: invalid length of keyrep
 *       -3: unknown decryption algorithm
 */
unsigned char* auth_decrypt(int alg, unsigned char *ssec, unsigned int sseclen,
			    struct msg_key *keyrep, struct reg_rep *data,
			    int *sk_len)
{
	unsigned char *skey;
	int i;
	int mac_len;
	MACfunc mac_func;

	mac_func = auth_get_mac_data(alg, &mac_len);
	if (mac_func == NULL) {
		DEBUG(DEBUG_FLAG, "auth_decrypt: unknown algorithm %i\n", alg);
		*sk_len = -3;
		return NULL;
	}

	*sk_len = mac_len;
	if (GET_KEY_LEN(keyrep) != *sk_len) {
		DEBUG(DEBUG_FLAG, "auth_decrypt: keyrep has wrong length %i "
		      "!= %i\n", GET_KEY_LEN(keyrep), *sk_len);
		*sk_len = -2;
		return NULL;
	}
	skey = malloc(*sk_len);
	if (skey == NULL) {
		DEBUG(DEBUG_FLAG, "auth_decrypt: not enough memory\n");
		*sk_len = -1;
		return NULL;
	}
	mac_func(ssec, sseclen, (unsigned char *) data, sizeof(struct reg_rep),
		 skey);
	for (i = 0; i < *sk_len; i++)
		skey[i] ^= *(MSG_KEY_DATA(keyrep) + i);
	return skey;
}


/* NOTE!
 * /dev/random waits until it gets enough "randomness" available
 * i.e. it can take a long time to generate the needed 128 bits!
 * the use of /dev/urandom eliminates this blocking, but it is not as
 * secure
 *
 * The "randomness" is coming from mouse movements, keyboard clicks,
 * interrupts and block device requests. If the system is very idle, there
 * might be some DoS problems with this.. Because of this, the system uses
 * /dev/urandom. This should be changed if a reliable non-blocking
 * source of random numbers was available.
 *
 * /dev/urandom gives as good random numbers as /dev/random _if_ there is
 * enough random data available, but it will not block in case such data
 * is not available (it gives some less random data in that case). The
 * session key is generated with MD5 and a secret key generation key, so the
 * process should be very difficult to break even if some of the random bits
 * were not so random.
 */
int generate_session_key(int alg, unsigned char *key, unsigned int key_len,
			 unsigned char *session_key)
{
	FILE *in;
	unsigned char data[MAX_SK_LEN];
	int len;
	int mac_len;
	MACfunc mac_func;

	mac_func = auth_get_mac_data(alg, &mac_len);
	if (mac_func == NULL) {
		DEBUG(DEBUG_FLAG,
		      "generate_session_key: unknown algorithm %i\n", alg);
		return -1;
	}

	len = mac_len;

	if (key == NULL)
		return -1;

	in = fopen("/dev/urandom", "rb");
	if (in == NULL)
		return -1;
	if (fread(data, 1, len, in) != len) {
		fclose(in);
		return -1;
	}
	fclose(in);

	mac_func(key, key_len, data, len, session_key);

	assert(len <= MAX_SK_LEN);
	return len;
}


/**
 * auth_is_protected:
 * @ext: pointer to extension whose protection is checked
 * @auth: pointer to authentication extension
 *
 * Checks whether @auth protects @ext, i.e., whether @auth comes after @ext.
 *
 * Returns:
 *   1 if @ext is %NULL or protected by @auth; or 0 if not
 */
int auth_is_protected(void *ext, void *auth)
{
	if (ext == NULL)
		return 1;
	if (ext > auth)
		return 0;
	return 1;
}


/**
 * auth_aaa_key_alg_id_to_dynamics:
 * @alg_id: algorithm id as specified in draft-ietf-mobileip-aaa-key-07.txt
 *
 * Convert algorithm id in aaa-key to authentication algorithm used in
 * Dynamics.
 *
 * Returns: 0 if auth. alg. is not recognized or corresponding Dynamics
 * authentication algorithm id.
 */
int auth_aaa_key_alg_id_to_dynamics(int alg_id)
{
	switch (alg_id) {
	case AAA_KEY_ALG_MD5:
		return AUTH_ALG_MD5;
	case AAA_KEY_ALG_HMAC_MD5:
		return AUTH_ALG_HMAC_MD5;
	case AAA_KEY_ALG_SHA1:
		return AUTH_ALG_SHA1;
	}

	return 0;
}


/**
 * auth_aaa_key_replay_to_dynamics:
 * @replay: replay method as specified in draft-ietf-mobileip-aaa-key-07.txt
 *
 * Convert replay method in aaa-key to method identifier used in Dynamics.
 *
 * Returns: -1 if auth. alg. is not recognized or corresponding Dynamics
 * replay method id.
 */
int auth_aaa_key_replay_to_dynamics(int replay)
{
	switch (replay) {
	case AAA_KEY_REPLAY_NONE:
		return REPLAY_PROTECTION_NONE;
	case AAA_KEY_REPLAY_TIMESTAMPS:
		return REPLAY_PROTECTION_TIMESTAMP;
	case AAA_KEY_REPLAY_NONCES:
		return REPLAY_PROTECTION_NONCE;
	}

	return 0;
}


/**
 * auth_generate_key:
 * @alg: AAA algorithm used for key generation
 * @key: AAA shared secret
 * @keylen: length of @key
 * @genkey: memory area for generated key
 * @genkeylen: pointer to @genkey len; initially maximum size and after
 * function call, length of the generated key
 *
 * Generate shared secret for dynamic security association as specified in
 * draft-ietf-mobile-aaa-key-07.txt.
 *
 * Returns: 0 on success or -1 on failure
 */
int auth_generate_key(int alg, unsigned char *key, unsigned int keylen,
		      const unsigned char *keymaterial,
		      unsigned int keymateriallen,
		      const unsigned char *nodeaddr, unsigned int nodeaddrlen,
		      unsigned char *genkey, unsigned int *genkeylen)
{
	unsigned char *data;
	int mac_len;
	MACfunc mac_func;

	mac_func = auth_get_mac_data(alg, &mac_len);
	if (mac_func == NULL) {
		DEBUG(DEBUG_FLAG, "auth_generate_key: unsupported algorithm "
		      "%i\n", alg);
		return -1;
	}

	if (*genkeylen < mac_len) {
		DEBUG(DEBUG_FLAG, "auth_generate_key: not enough space"
		      " for generated key\n");
		return -1;
	}

	data = (unsigned char *) malloc(keymateriallen + nodeaddrlen);
	if (data == NULL) {
		DEBUG(DEBUG_FLAG, "auth_generate_key: could not allocate "
		      "memory\n");
		return -1;
	}

	memcpy(data, keymaterial, keymateriallen);
	memcpy(data + keymateriallen, nodeaddr, nodeaddrlen);

	mac_func(key, keylen, data, keymateriallen + nodeaddrlen, genkey);
	*genkeylen = mac_len;
	free(data);

	return 0;
}


/**
 * auth_supported_auth_alg:
 * @alg: algorithm id
 *
 * Check whether given algorithm is supported for authentication.
 *
 * Returns: 1 if supported, or 0 if not
 */
int auth_supported_auth_alg(int alg)
{
	if (alg == AUTH_ALG_MD5 || alg == AUTH_ALG_MD5_RFC2002 ||
	    /* AUTH_ALG_RADIUS not supported for all functions */
	    alg == AUTH_ALG_MD5_PREFIX_SUFFIX ||
	    alg == AUTH_ALG_HMAC_MD5 || alg == AUTH_ALG_SHA1 ||
	    alg == AUTH_ALG_HMAC_SHA1)
		return 1;

	return 0;
}
