/* $Id: fileio.c,v 1.25 2001/07/26 17:31:57 jm Exp $
 * Configuration file I/O functions
 *
 * Dynamic hierarchial IP tunnel
 * Copyright (C) 1998-2001, Dynamics group
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation. See README and COPYING for
 * more details.
 */

#include "owntypes.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <assert.h>

#include <sys/types.h> /* required for netinet/in.h in AMIGAOS */
#include <netinet/in.h>
#include <arpa/inet.h>

#include <syslog.h>

#include "fileio.h"
#include "message.h"
#include "dyn_ip.h"
#include "util.h"

#ifndef LOG_FTP
#define LOG_FTP         (11<<3) /* ftp daemon */
#endif

/* defines */

#define ASSERT assert

#define BUFLEN 128                   /* input buffer length to start with */

#define DEBUGSAVE

#define IP_ADDR_LEN 4

/* Load data from an ASCII file
 *
 * Gets the keyword and data data from each line and calls a user-specified
 * function to process the line
 *
 * INPUT VALUES:
 *
 *    ) Userdata may be anything, just passed to the processing function
 *    ) File must not be NULL, internal error, ASSERT used
 * M01) Not enough memory for buffer
 * M02) Blank lines
 * M03) Comment in the beginning of the line
 * M04) Comment not in the beginning of the line
 * M05) Configuration entry before a comment
 * M06) The line is longer than the default buffer size
 * M07) Spaces before the keyword
 *
 */

int load_data(void *userdata, FILE *file,
	      int (*process) (void *, char *, char *))
{
	int i, result, len, flag, bufsize;
	char *buf, *key, *data, *temp;

	ASSERT(file != NULL);
	bufsize = BUFLEN;
	buf = malloc(bufsize);
	if (buf == NULL) {
		fprintf(stderr,
			"load_data: Not enough memory for buffer "
			"(%d bytes)!\n", bufsize);
		return FALSE;
	}
	do {
		if (fgets(buf, bufsize, file) == NULL) break;
		len = strlen(buf);
		ASSERT(len <= bufsize - 1);
		while (buf[len - 1] != '\n') {
			bufsize *= 2;
			temp = realloc(buf, bufsize);
			if (temp == NULL) {
				fprintf(stderr,
					"load_data: Not enough memory for "
					"buffer (%d bytes)!\n", bufsize);
				free(buf);
				return FALSE;
			}
			buf = temp;
			if (fgets(buf + len, bufsize - len, file) == NULL) {
				break;
			}
			len = strlen(buf);
			ASSERT(len <= bufsize - 1);
			if (len < bufsize - 1 && buf[len - 1] != '\n') {
				fprintf(stderr, "load_data: Binary file "
					"containing string terminator or too "
					"long line\n");
				free(buf);
				return FALSE;
			}
		}
		if (buf[len - 1] != '\n') {
			fprintf(stderr,
				"load_data: Line is not terminated with a "
				"linefeed\n");
			free(buf);
			return FALSE;
		}
		buf[len - 1] = '\0'; /* remove trailing newline */
		data = NULL;
		flag = FALSE;
		for (i = 0; i < strlen(buf); i++) {
			if (buf[i] != ' ') break;
		}
		key = buf + i;
		for (i = 0; i < strlen(key); i++) {
			if (key[i] == '"') {
				flag = !flag;
			}
			if (key[i] == '#' && flag == FALSE) {
				key[i] = '\0';
				break;
			}
		}
		for (i = 0; i < strlen(key); i++) {
			if (key[i] == ' ' || key[i] == '\t') {
				key[i] = '\0';
				data = key + i + 1;
				break;
			}
		}
		if (strlen(key) != 0) {
			result = (*process)(userdata, key, data);
			if (result < 0) {
				fprintf(stderr,
					"load_data: key = %s, data = %s : ",
					key, data);
				fprintf(stderr, "result = %d\n", result);
				break;
			}
			if (result > 0) {
				free(buf);
				return TRUE;
			}
		}
	} while (1);
	free(buf);
	return FALSE;
}

/* Load integer
 *
 * Integers starting with:
 * - "0x" are considered to be hexadecimal.
 * - "0" are considered to be octal.
 * - anything else is considered to be decimal.
 *
 * INPUT VALUES:
 *
 *    ) data must not be NULL
 *    ) x must not be NULL
 * M08) Hexadecimal number
 * M09) Octal number
 * M10) Decimal number
 * M11) Something else
 */

int load_int(const char *data, int *x)
{
	const char *pos = data;

	ASSERT(data != NULL);
	ASSERT(x != NULL);

	*x = 0;

	while (*pos == ' ' || *pos == '\t')
		pos++;

	if (strncmp(pos, "0x", 2) == 0) {
		if (sscanf(pos, "%x", x) != 1) return FALSE;
	} else if (strncmp(pos, "0", 1) == 0) {
		if (sscanf(pos, "%o", x) != 1) return FALSE;
	} else {
		if (sscanf(pos, "%d", x) != 1) return FALSE;
	}

	return TRUE;
}

/* Load boolean
 *
 * Boolean can be either "TRUE" or "FALSE"
 *
 * INPUT VALUES:
 *
 *    ) data must not be NULL
 *    ) x must not be NULL
 * M12) TRUE
 * M13) FALSE
 * M14) Something else
 */

int load_bool(const char *data, int *x)
{
	const char *pos = data;

	ASSERT(data != NULL);
	ASSERT(x != NULL);
	*x = FALSE;

	while (*pos == ' ' || *pos == '\t')
		pos++;

	if (strcmp(pos, "TRUE") == 0) {
		*x = TRUE;
		return TRUE;
	}
	if (strcmp(pos, "FALSE") == 0) {
		return TRUE;
	}
	return FALSE;
}

/* Load ip address
 *
 * INPUT VALUES:
 *
 * IP address is in the IPv4 format, a.b.c.d
 *
 *    ) data must not be NULL
 *    ) addr must not be NULL
 * M15) Valid IP address
 * M16) IP byte is not a number
 * M17) IP byte higher than 255
 * M18) IP byte lower than 0
 * M19) No separating dot
 * M20) Too few IP bytes
 * M21) Too many IP bytes
 * M22) Something else
 */

int load_ip_address(const char *data, struct in_addr *addr)
{
	int i, tmp;
	unsigned int ip;
	const char *pos;

	ASSERT(data != NULL);
	ASSERT(addr != NULL);
	pos = data;
	while (*pos == ' ' || *pos == '\t')
		pos++;
	ip = 0;
	for (i = 0; ; i++) {
		if (sscanf(pos, "%d", &tmp) != 1) {
			fprintf(stderr,
				"load_ip_address: Unable to parse IP byte "
				"from string '%s'\n", pos);
			return FALSE;
		}
		if (tmp > 255) {
			fprintf(stderr,
				"load_ip_address: Invalid IP byte %d higher "
				"than 255\n", tmp);
			return FALSE;
		}
		if (tmp < 0) {
			fprintf(stderr,
				"load_ip_address: Invalid IP byte %d lower "
				"than 0\n", tmp);
			return FALSE;
		}
		ASSERT(tmp >= 0);
		ASSERT(tmp <= 255);
		ip = (ip << 8) | tmp;
		pos = pos + strspn(pos, "0123456789");
		if (*pos == '\0' || *pos == ' ' || *pos == '\t') {
			if (i == IP_ADDR_LEN - 1) break;
			fprintf(stderr,
				"load_ip_address: Separating dot not found\n");
			return FALSE;
		}
		if (*pos == '.') {
			if (i == IP_ADDR_LEN - 1) {
				fprintf(stderr,
					"load_ip_address: Too many IP "
					"bytes, only %u expected\n",
					IP_ADDR_LEN);
				return FALSE;
			}
			pos++;
		} else {
			fprintf(stderr,
				"load_ip_address: IP byte is not a number\n");
			return FALSE;
		}
	}
	addr->s_addr = htonl(ip);
	return TRUE;
}

/**
 * load_net_address:
 * @char: Any IP address with or without a mask. Valid input string is any
 *        of the following: 
 *          a.b.c.d  
 *          a.b.c.d/p.q.r.s
 *          a.b.c.d/n   (0 <= n <= 32)
 * @addr:
 * @mask:
 */
int load_net_address(char *data, struct in_addr *addr,
		     struct in_addr *mask)
{
	char *pos;
	int tmp;
	int masktable[33] = { 0x00000000,
			      0x80000000, 0xc0000000, 0xe0000000, 0xf0000000,
			      0xf8000000, 0xfc000000, 0xfe000000, 0xff000000,
			      0xff800000, 0xffc00000, 0xffe00000, 0xfff00000,
			      0xfff80000, 0xfffc0000, 0xfffe0000, 0xffff0000,
			      0xffff8000, 0xffffc000, 0xffffe000, 0xfffff000,
			      0xfffff800, 0xfffffc00, 0xfffffe00, 0xffffff00,
			      0xffffff80, 0xffffffc0, 0xffffffe0, 0xfffffff0,
			      0xfffffff8, 0xfffffffc, 0xfffffffe, 0xffffffff,
	};

	pos = strchr(data, '/');
	if (pos == NULL) {
		if (load_ip_address(data, addr) != TRUE) {
			fprintf(stderr,
				"load_net_address: invalid IP address\n");
			return FALSE;
		}
		mask->s_addr = htonl(0xffffffff);
	} else {
		*pos = '\0';
		if (load_ip_address(data, addr) != TRUE) {
			fprintf(stderr,
				"load_net_address: invalid IP address\n");
			return FALSE;
		}
		pos++;
		if (strchr(pos, '.') != NULL) {
			if (load_ip_address(pos, mask) != TRUE) {
				fprintf(stderr,
				"load_net_address: invalid network mask\n");
				return FALSE;
			}
		} else {
			if (sscanf(pos, "%d", &tmp) == 1 &&
			    tmp >= 0 && tmp <= 32) {
				ASSERT(tmp >= 0);
				ASSERT(tmp <= 32);
				mask->s_addr = htonl(masktable[tmp]);
			} else {
				fprintf(stderr,
					"load_net_address: invalid network "
					"mask\n");
				return FALSE;
			}
		}

	}
	return TRUE;
}

/**
 * load_ip_prefix:
 * @char: Any IP address with or without a mask. Valid input string is any
 *        of the following: 
 *          a.b.c.d  
 *          a.b.c.d/n   (0 <= n <= 32)
 * @addr:
 * @prefix_len:
 */
int load_ip_prefix(char *data, struct in_addr *addr, int *prefix_len)
{
	char *pos;

	pos = strchr(data, '/');
	if (pos == NULL) {
		if (load_ip_address(data, addr) != TRUE) {
			fprintf(stderr,
				"load_net_address: invalid IP address\n");
			return FALSE;
		}
		*prefix_len = 32;
	} else {
		*pos = '\0';
		if (load_ip_address(data, addr) != TRUE) {
			fprintf(stderr,
				"load_ip_prefix: invalid IP address\n");
			return FALSE;
		}
		pos++;
		if (!(sscanf(pos, "%d", prefix_len) == 1 &&
		    *prefix_len >= 0 && *prefix_len <= 32)) {
			fprintf(stderr,
				"load_ip_prefix: invalid network prefix\n");
			return FALSE;
		}
	}
	return TRUE;
}

/* Load hex table
 *
 * Reads hex table into the array and optionally sets the length read
 *
 * INPUT VALUES:
 *
 *    ) data must not be NULL
 *    ) table must not be NULL
 * M23) Valid hex table
 * M24) Invalid odd character
 * M25) Invalid even character
 * M26) Incomplete hex byte
 * M27) Too long hex value
 * M28) Length is NULL
 * M29) Something else
 */

int load_hex_table(const char *data, unsigned char *table, int maxlen,
		   int *length)
{
	int i, tmp, res;
	char c;

	ASSERT(data != NULL);
	ASSERT(table != NULL);
	if (length != NULL) {
		*length = 0;
	}
	for (i = 0; i < maxlen; i++) {
		table[i] = '\0';
	}
	while (*data == ' ' || *data == '\t') data++;
	if (strncmp(data, "0x", 2) == 0) {
		data += 2;
	} else if (*data == '"') {
		res = load_char_table(data, (char *)table, maxlen);
		if (length != NULL && res == TRUE) {
			table[maxlen - 1] = '\0';
			*length = strlen((char *)table);
		}
		return res;
	}
	for (i = 0; i < maxlen; i++) {
		do {
			c = *data++;
		} while (c == ' ' || c == '\t');
		if (c == '\0') break;
		tmp = get_hex_digit(c);
		if (tmp == -1) {
			fprintf(stderr,
				"load_hex_table: invalid character '%c'\n",
				c);
			return FALSE;
		}
		res = tmp * 16;
		c = *data++;        /* no spaces allowed in one hex byte */
		if (c == '\0') {
			fprintf(stderr,
				"load_hex_table: incomplete hex byte\n");
			return FALSE;
		}
		tmp = get_hex_digit(c);
		if (tmp == -1) {
			fprintf(stderr,
				"load_hex_table: invalid character '%c'\n",
				c);
			return FALSE;
		}
		res += tmp;
		table[i] = res;
	}
	if (i == maxlen) {
		do {
			c = *data++;
		} while (c == ' ' || c == '\t');
		if (c != '\0') {
			fprintf(stderr,
				"load_hex_table: too long hex value\n");
			return FALSE;
		}
	}
	if (length != NULL) {
		*length = i;
	}
	return TRUE;
}

/* Load character table (string)
 *
 * NOTE! The routine loads actually a table of characters.
 * If the string is as long as the buffer, the string will not be
 * null-terminated! The user of the routine has to make sure that the
 * string is null-terminated if required.
 *
 * INPUT VALUES:
 *
 *    ) data must not be NULL
 *    ) table must not be NULL
 * M30) Valid string
 * M31) 2 x '"' in string
 * M32) only one '"' in string
 * M33) '"' in the middle string
 * M34) '\n' in string
 * M35) '\r' in string
 * M36) '\\' in string
 * M37) '\"' in string
 * M38) '\oct' in string
 * M39) Too long string
 */

int load_char_table(const char *data, char *table, int maxlen)
{
	int i, j, startpos;
	char temp[5];
	int val, write, len;
	char ch;
	int start, stop;

	ASSERT(data != NULL);
	ASSERT(table != NULL);
	for (i = 0 ; i < maxlen ; i++) {
		table[i] = '\0';
	}
	write = 0;
	start = FALSE;
	stop = FALSE;
	startpos = 0;
	while (data[startpos] == ' ' || data[startpos] == '\t')
		startpos++;
	len = strlen(data) - startpos;
	for (i = startpos; i <= len; i++) {
		val = data[i];
		if (val == '\0') {
			if (start == TRUE && stop == TRUE) break;
			fprintf(stderr,
				"load_char_table: Unexpected end of string\n");
			return FALSE;
		}
		if (val == '"') {
			if (start == FALSE) {
				start = TRUE;
				continue;
			} else if (stop == FALSE) {
				stop = TRUE;
				continue;
			} else {
				fprintf(stderr,
					"load_char_table: Unexpected '\"'\n");
				return FALSE;
			}
		} else {
			if (start == FALSE) {
				fprintf(stderr,
					"load_char_table: Unrecognized "
					"character '%c'\n",
					val);
				return FALSE;
			}
		}
		if (val == '\\') {
			if (i == len - 1) {
				fprintf(stderr,
					"load_char_table: String longer "
					"than expected\n");
				return FALSE;
			}
			i++;
			if (data[i] == '0') {
				ch = '\0';
			} else if (data[i] == 'n') {
				ch = '\n';
			} else if (data[i] == 'r') {
				ch = '\r';
			} else if (data[i] == '\\') {
				ch = '\\';
			} else if (data[i] == '"') {
				ch = '"';
			} else {
				if (i >= len - 3) {
					fprintf(stderr,
						"load_char_table: String "
						"longer than expected\n");
					return FALSE;
				}
				for (j = 0 ; j < 3 ; j++) {
					ch = data[i];
					i++;
					if (ch < '0' || ch > '9') {
						fprintf(stderr,
							"load_char_table: "
							"Invalid octal "
							"number\n");
						return FALSE;
					}
					temp[j] = ch;
				}
				i--;
				temp[3] = '\0';
				if (sscanf(temp, "%03o", &val) != 1) {
					fprintf(stderr,
						"load_char_table: Invalid "
						"octal number\n");
					return FALSE;
				}
				ch = val;
			}
		} else {
			ch = val;
		}
		if (write == maxlen) {
			fprintf(stderr,
				"load_char_table: String longer than "
				"expected\n");
			return FALSE;
		}
		table[write] = ch;
		write++;
	}
	if (write < maxlen) {
		table[write] = '\0';
	}
	return TRUE;
}

/* Load syslog facility
 *
 * INPUT VALUES:
 *
 *    ) data must not be NULL
 * M40) Known syslog facility
 * M41) Unknown syslog facility
 */

int load_syslog_facility(const char *data, int *facility)
{
	struct facilitydata {
		char *name;
		int code;
	};
	struct facilitydata facilitycode[] = {
		{ "LOG_KERN",     LOG_KERN     },
		{ "LOG_MAIL",     LOG_MAIL     },
		{ "LOG_DAEMON",   LOG_DAEMON   },
		{ "LOG_AUTH",     LOG_AUTH     },
		{ "LOG_SYSLOG",   LOG_SYSLOG   },
		{ "LOG_LPR",      LOG_LPR      },
		{ "LOG_NEWS",     LOG_NEWS     },
		{ "LOG_UUCP",     LOG_UUCP     },
		{ "LOG_CRON",     LOG_CRON     },
		{ "LOG_AUTHPRIV", LOG_AUTHPRIV },
		{ "LOG_FTP",      LOG_FTP      },
		{ "LOG_LOCAL0",   LOG_LOCAL0   },
		{ "LOG_LOCAL1",   LOG_LOCAL1   },
		{ "LOG_LOCAL2",   LOG_LOCAL2   },
		{ "LOG_LOCAL3",   LOG_LOCAL3   },
		{ "LOG_LOCAL4",   LOG_LOCAL4   },
		{ "LOG_LOCAL5",   LOG_LOCAL5   },
		{ "LOG_LOCAL6",   LOG_LOCAL6   },
		{ "LOG_LOCAL7",   LOG_LOCAL7   },
		{ NULL,           -1           }
	};

	int i;

	for (i = 0; facilitycode[i].name != NULL; i++) {
		if (strcmp(data, facilitycode[i].name) == 0) {
			*facility = facilitycode[i].code;
			return TRUE;
		}
	}
	return FALSE;
}


int load_nai(const char *data, char *nai, int *len)
{
	char *start, *end, *t;
	char hwbuf[128], tmp[MAX_NAI_LEN];

	if (load_char_table(data, nai, MAX_NAI_LEN) != TRUE)
		return FALSE;

	start = strchr(nai, '[');
	if (start != NULL) {
		end = strchr(start, ']');
		if (end == NULL || end < start) {
			fprintf(stderr,
				"load_nai - parse error ('[' without ']')\n");
			return FALSE;
		}
		*end = '\0';
		if (dyn_ip_get_hwaddr((start + 1), hwbuf, sizeof(hwbuf)) < 0) {
			*end = ']';
			fprintf(stderr,	"load_nai - cannot resolve hwaddr for "
				"interface '%s'.\n",
				(start + 1));
			return FALSE;
		}
		t = hwbuf;
		while (*t != '\0') {
			if (*t == ':')
				*t = '.';
			t++;
		}
		if (strlen(nai) + strlen(hwbuf) + start - end - 1 >=
		    MAX_NAI_LEN) {
			fprintf(stderr, "load_nai - too long NAI\n");
			return FALSE;
		}
		dynamics_strlcpy(tmp, end + 1, sizeof(tmp));
		dynamics_strlcpy(start, hwbuf, MAX_NAI_LEN - (start - nai));
		dynamics_strlcpy(start + strlen(hwbuf), tmp,
				 MAX_NAI_LEN - (start + strlen(hwbuf) - nai));
	}

	*len = strlen(nai);

	return TRUE;
}
